#!/usr/bin/env python3

import os
import glob
import subprocess as sp
import logging

logging.basicConfig(level=logging.INFO)

OUTPUT_PATH = 'packages'
REPLACEMENTS = {
    'Submodules\n----------\n\n': '',
    'Subpackages\n-----------': '**Subpackages:**',
    'pde package\n===========': 'Reference manual\n================',
}



def replace_in_file(infile, replacements, outfile=None):
    """ reads in a file, replaces the given data using python formatting and
    writes back the result to a file.
    
    Args:
        infile (str):
            File to be read
        replacements (dict):
            The replacements old => new in a dictionary format {old: new}
        outfile (str):
            Output file to which the data is written. If it is omitted, the
            input file will be overwritten instead
        
    """
    if outfile is None:
        outfile = infile
    
    with open(infile, 'r') as fp:
        content = fp.read()
        
    for key, value in replacements.items():
        content = content.replace(key, value)
    
    with open(outfile, 'w') as fp:
        fp.write(content)    



def main():
    # remove old files
    for path in glob.glob(f'{OUTPUT_PATH}/*.rst'):
        logging.info('Remove file `%s`', path)
        os.remove(path)
    
    # run sphinx-apidoc
    sp.check_call(['sphinx-apidoc',
                   '--maxdepth', '4',
                   '--output-dir', OUTPUT_PATH,
                   '--module-first',
                   '../../pde',  # path of the package
                   '../../pde/tests',  # ignored path
                   '../../pde/**/tests'  # ignored path
                   ])

    # replace unwanted information
    for path in glob.glob(f'{OUTPUT_PATH}/*.rst'):
        logging.info('Patch file `%s`', path)
        replace_in_file(path, REPLACEMENTS)

    
    
if __name__ == '__main__':
    main()



"""
Handling mathematical expressions with sympy

This module provides classes representing expressions that can be provided as 
human-readable strings and are converted to :mod:`numpy` and :mod:`numba`
representations using :mod:`sympy`.

.. autosummary::
   :nosignatures:

   parse_number
   ScalarExpression
   TensorExpression
   
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de> 
"""


from abc import ABCMeta, abstractproperty
import copy
import logging
import re
from typing import (Union, Callable, List, Optional, Dict, Any,
                    Set, Sequence, Tuple)  # @UnusedImport
from numbers import Number

import sympy
import numpy as np

from .cache import cached_property, cached_method
from .docstrings import fill_in_docstring

from .numba import jit



@fill_in_docstring
def parse_number(expression: Union[str, float],
                 variables: Dict[str, float] = None) -> float:
    r""" return a number compiled from an expression
    
    Warning:
        {WARNING_EXEC}
    
    Args:
        expression (str or float):
            An expression that can be interpreted as a number
        variables (dict):
            A dictionary of values that replace variables in the expression
        
    Returns:
        float: the calculated value
    """
    from sympy.parsing import sympy_parser
    
    if variables is None:
        variables = {}
    
    expr = sympy_parser.parse_expr(str(expression))
    try:
        value = float(expr.subs(variables))
    except TypeError as err:
        if not err.args: 
            err.args = ('',)
        err.args = err.args + (f"Expression: `{expr}`",)
        raise 
    return value



ExpressionType = Union[float, str, "ExpressionBase"]


class ExpressionBase(metaclass=ABCMeta):
    """ abstract base class for handling expressions """

    @fill_in_docstring
    def __init__(self, expression,
                 signature: Optional[Sequence[Union[str, List[str]]]] = None,
                 user_funcs: Optional[Dict[str, Any]] = None):
        """
        Warning:
            {WARNING_EXEC}
    
        Args:
            expression (str or float):
                The expression, which is either a number or a string that sympy
                can parse
            signature (list of str):
                The signature defines which variables are expected in the
                expression. This is typically a list of strings identifying
                the variable names. Individual names can be specified as list,
                in which case any of these names can be used. The first item in
                such a list is the definite name and if another name of the list
                is used, the associated variable is renamed to the definite
                name. If signature is `None`, all variables in `expressions`
                are allowed.
            user_funcs (dict, optional):
                A dictionary with user defined functions that can be used in the
                expression
        """
        self._sympy_expr = expression
        self._logger = logging.getLogger(self.__class__.__name__)
        self.user_funcs = {} if user_funcs is None else user_funcs
        self._check_signature(signature)


    def __repr__(self):
        return (f'{self.__class__.__name__}("{self.expression}", '
                f'signature={self.vars})')
        
        
    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        # compare what the expressions depend on
        if set(self.vars) != set(other.vars):
            return False
        
        # compare the expressions themselves by checking their difference
        diff = sympy.simplify(self._sympy_expr - other._sympy_expr)
        if isinstance(self._sympy_expr, sympy.NDimArray):
            return diff == sympy.Array(np.zeros(self._sympy_expr.shape))
        else:
            return diff == 0


    @property
    def constant(self) -> bool:
        """ bool: whether the expression is a constant """
        return len(self._sympy_expr.free_symbols) == 0


    @abstractproperty
    def shape(self) -> Tuple[int, ...]: pass


    def _check_signature(self,
                         signature: Sequence[Union[str, List[str]]] = None):
        """ validate the variables of the expression against the signature """
        # get arguments of the expressions
        if self.constant:
            # constant expression do not depend on any variables
            args: Set[str] = set()
            if signature is None:
                signature = []
                
        else:
            # general expressions might have a variable
            args = set(str(s).split('[')[0]
                       for s in self._sympy_expr.free_symbols)
            if signature is None:
                # create signature from arguments
                signature = list(sorted(args))
        
        self._logger.debug(f'Expression arguments: {args}')
        
        # check whether variables are in signature
        self.vars: Any = []
        found = set()
        for sig in signature:
            sig_list = [sig] if isinstance(sig, str) else sig
                
            # use the first item as the variable name
            arg_name = sig_list[0]
            self.vars.append(arg_name)

            # check whether this part of the signature is present                 
            for arg in args:
                if arg in sig_list:
                    if arg != arg_name:  # synonym has been used
                        old = sympy.symbols(arg)
                        new = sympy.symbols(arg_name)
                        self._sympy_expr = self._sympy_expr.subs(old, new)
                        self._logger.info(f'Renamed variable "{old}"->"{new}"')
                    found.add(arg)
                    break

        args = set(args) - found 
        if len(args) > 0:
            raise RuntimeError(f'Arguments {args} were not defined in '
                               f'expression signature {signature}')
            
    
    @property
    def expression(self) -> str:
        """ str: the expression in string form """
        # turn numerical values into easily readable text
        if isinstance(self._sympy_expr, sympy.NDimArray):
            expr = self._sympy_expr.applyfunc(lambda x: x.evalf(chop=True))
        else:
            expr = self._sympy_expr.evalf(chop=True)
            
        return str(expr.xreplace({n: float(n)
                                  for n in expr.atoms(sympy.Float)}))

        
    @property
    def rank(self) -> int:
        """ int: the rank of the expression """
        return len(self.shape)
        
        
    def depends_on(self, variable: str) -> bool:
        """ determine whether the expression depends on `variable`
        
        Args:
            variable (str): the name of the variable to check for
            
        Returns:
            bool: whether the variable appears in the expression
        """
        if self.constant:
            return False
        else:
            return any(variable == str(symbol) 
                       for symbol in self._sympy_expr.free_symbols) 
    
    
    def _get_function(self, single_arg: bool = False,
                      user_funcs: Dict[str, Callable] = None) -> Callable:
        """ return function evaluating expression
        
        Args:
            single_arg (bool):
                Determines whether the returned function accepts all variables
                in a single argument as an array or whether all variables need
                to be supplied separately
            user_funcs (dict):
                Additional functions that can be used in the expression 
        
        Returns:
            function: the function
        """
        if user_funcs is None:
            user_funcs = {}
        
        variables = (self.vars,) if single_arg else self.vars
        return sympy.lambdify(variables, self._sympy_expr,  # type: ignore
                              modules=[user_funcs, self.user_funcs, 'numpy'])
        
        
    @cached_method()
    def _get_function_cached(self, single_arg: bool = False) -> Callable:
        """ return function evaluating expression
        
        Args:
            single_arg (bool):
                Determines whether the returned function accepts all variables
                in a single argument as an array or whether all variables need
                to be supplied separately
        
        Returns:
            function: the function
        """
        return self._get_function(single_arg)


    def __call__(self, *args, **kwargs):
        """ return the value of the expression for the given values """
        return self._get_function_cached(single_arg=False)(*args, **kwargs)

    
    @cached_method()
    def get_compiled(self, single_arg: bool = False) -> Callable:  
        """ return numba function evaluating expression
        
        Args:
            single_arg (bool): Determines whether the returned function accepts
                all variables in a single argument as an array or whether all
                variables need to be supplied separately
        
        Returns:
            function: the compiled function
        """
        func = self._get_function_cached(single_arg=single_arg)
        return jit(func)  # type: ignore
    


class ScalarExpression(ExpressionBase):
    """ describes a mathematical expression of a scalar quantity """
    
    shape: Tuple[int, ...] = tuple()
    
    
    @fill_in_docstring
    def __init__(self, expression: ExpressionType = 0,
                 signature: Optional[Sequence[Union[str, List[str]]]] = None,
                 user_funcs: Optional[Dict[str, Any]] = None,
                 allow_indexed: bool = False):
        """
        Warning:
            {WARNING_EXEC}

        Args:
            expression (str or float):
                The expression, which is either a number or a string that sympy
                can parse
            signature (list of str):
                The signature defines which variables are expected in the
                expression. This is typically a list of strings identifying
                the variable names. Individual names can be specified as lists,
                in which case any of these names can be used. The firstm item in
                such a list is the definite name and if another name of the list
                is used, the associated variable is renamed to the definite
                name. If signature is `None`, all variables in `expressions`
                are allowed.
            user_funcs (dict, optional):
                A dictionary with user defined functions that can be used in the
                expression
            allow_indexed (bool):
                Whether to allow indexing of variables. If enabled, array
                variables are allowed to be indexed using square bracket
                notation.
        """
        self.allow_indexed = allow_indexed
        
        # parse the expression
        if isinstance(expression, ScalarExpression):
            # copy constructor
            sympy_expr = copy.copy(expression._sympy_expr)
            signature = expression.vars
            self.allow_indexed = expression.allow_indexed
            if user_funcs is None:
                user_funcs = expression.user_funcs
            else:
                user_funcs.update(expression.user_funcs)
            
        elif callable(expression):
            # expression is some other callable -> not allowed anymore
            raise TypeError('Expression must be provided as string and not as '
                            'a callable function')
        
        elif isinstance(expression, Number):
            # expression is a simple number
            sympy_expr = sympy.Float(expression)
            
        elif bool(expression):
            # parse expression as a string
            expression = self._prepare_expression(str(expression))

            # parse the expression using sympy                
            from sympy.parsing import sympy_parser
            sympy_expr = sympy_parser.parse_expr(expression)
            
        else:
            # expression is empty, False or None => set it to zero                
            sympy_expr = sympy.Float(0)
            
        super().__init__(expression=sympy_expr, signature=signature,
                         user_funcs=user_funcs)
    
    
    @property
    def value(self) -> float:
        """ float: the value for a constant expression """
        if self.constant:
            return float(self._sympy_expr)
        else:
            raise TypeError('Only constant expressions have a defined value')
        
        
    @property
    def is_zero(self) -> bool:
        """ bool: returns whether the expression is zero """
        return self.constant and self.value == 0
        
        
    def __bool__(self):
        """ tests whether the expression is nonzero """
        return not self.constant or self.value != 0
        
        
    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (super().__eq__(other) and
                self.allow_indexed == other.allow_indexed)
        

    def _prepare_expression(self, expression: str) -> str:
        """ replace indexed variables, if allowed """
        if self.allow_indexed:
            return re.sub(r"(\w+)(\[\w+\])", r"IndexedBase(\1)\2", expression)
        else:
            return expression
        
        
    def _var_indexed(self, var: str) -> bool:
        """ checks whether the variable `var` is used in an indexed form """
        from sympy.tensor.indexed import Indexed
        return any(isinstance(s, Indexed) and s.base.name == var
                   for s in self._sympy_expr.free_symbols)


    def differentiate(self, var: str) -> "ScalarExpression":
        """ return the expression differentiated with respect to var """
        if self.constant:
            # return empty expression
            return ScalarExpression(expression=0,
                                    signature=self.vars,
                                    allow_indexed=self.allow_indexed)  
        if self.allow_indexed:
            if self._var_indexed(var):
                # TODO: implement this
                raise NotImplementedError('Cannot differentiate with respect '
                                          'to a vector')
            
        var = self._prepare_expression(var)
        return ScalarExpression(self._sympy_expr.diff(var),
                                signature=self.vars,
                                allow_indexed=self.allow_indexed,
                                user_funcs=self.user_funcs)
    
    
    @cached_property()
    def derivatives(self) -> "TensorExpression":
        """ differentiate the expression with respect to all variables """ 
        if self.constant:
            # return empty expression
            value = np.zeros(len(self.vars))
            return TensorExpression(expression=value,
                                    signature=self.vars)
            
        if self.allow_indexed:
            if any(self._var_indexed(var) for var in self.vars):
                raise RuntimeError('Cannot calculate gradient for expressions '
                                   'with indexed variables')
                
        grad = sympy.Array([self._sympy_expr.diff(v) for v in self.vars])
        return TensorExpression(sympy.simplify(grad),
                                signature=self.vars,
                                user_funcs=self.user_funcs)
        
        
        
class TensorExpression(ExpressionBase):
    """ describes a mathematical expression of a tensorial quantity """
    
    
    @fill_in_docstring
    def __init__(self, expression: ExpressionType,
                 signature: Optional[Sequence[Union[str, List[str]]]] = None,
                 user_funcs: Optional[Dict[str, Any]] = None):
        """
        Warning:
            {WARNING_EXEC}

        Args:
            expression (str or float):
                The expression, which is either a number or a string that sympy
                can parse
            signature (list of str):
                The signature defines which variables are expected in the
                expression. This is typically a list of strings identifying
                the variable names. Individual names can be specified as list,
                in which case any of these names can be used. The first item in
                such a list is the definite name and if another name of the list
                is used, the associated variable is renamed to the definite
                name. If signature is `None`, all variables in `expressions`
                are allowed.
            user_funcs (dict, optional):
                A dictionary with user defined functions that can be used in the
                expression
        """
        from sympy.tensor.array.ndim_array import ImmutableNDimArray

        # parse the expression
        if isinstance(expression, TensorExpression):
            # copy constructor
            sympy_expr = copy.copy(expression._sympy_expr)
            if user_funcs is None:
                user_funcs = expression.user_funcs
            else:
                user_funcs.update(expression.user_funcs)
            
        elif isinstance(expression, np.ndarray):
            # expression is a constant array
            sympy_expr = sympy.sympify(expression)
            
        elif isinstance(expression, ImmutableNDimArray):
            # expression is an array of sympy expressions
            sympy_expr = expression
            
        else:
            # parse expression as a string
            expression = str(expression)

            # parse the expression using sympy                
            from sympy.parsing import sympy_parser
            parsed = sympy.Array(sympy_parser.parse_expr(expression))
            sympy_expr = sympy.simplify(parsed)

        super().__init__(expression=sympy_expr, signature=signature,
                         user_funcs=user_funcs)


    @property
    def shape(self) -> Tuple[int, ...]:
        """ tuple: the shape of the tensor """
        return self._sympy_expr.shape  # type: ignore


    @property
    def value(self):
        """ the value for a constant expression """
        if self.constant:
            return np.array(self._sympy_expr.tolist(), dtype=np.double)
        else:
            raise TypeError('Only constant expressions have a defined value')

        
    def differentiate(self, var: str) -> "TensorExpression":
        """ return the expression differentiated with respect to var """
        if self.constant:
            # return empty expression
            return TensorExpression(expression=np.zeros(self.shape),
                                    signature=self.vars)
        return TensorExpression(self._sympy_expr.diff(var),
                                signature=self.vars,
                                user_funcs=self.user_funcs)
    
    
    @cached_property()
    def derivatives(self) -> "TensorExpression":
        """ differentiate the expression with respect to all variables """ 
        shape = (len(self.vars),) + self.shape
        
        if self.constant:
            # return empty expression
            return TensorExpression(np.zeros(shape), signature=self.vars)
                
        # perform the derivatives with respect to all variables
        derivs = sympy.derive_by_array(self._sympy_expr,
                                       [sympy.Symbol(s) for s in self.vars])
        return TensorExpression(derivs,
                                signature=self.vars,
                                user_funcs=self.user_funcs)
            
            
            
__all__ = ["ExpressionBase", "ScalarExpression", "TensorExpression"]



'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import itertools
import pickle

import pytest

from ..parameters import (Parameter, DeprecatedParameter, HideParameter,
                          Parameterized, get_all_parameters)



def test_parameters():
    """ test mixing Parameterized """
        
    param = Parameter('a', 1, int, "help")
    assert isinstance(str(param), str)
    
    p_string = pickle.dumps(param)
    param_new = pickle.loads(p_string)
    assert param.__dict__ == param_new.__dict__
    assert param is not param_new
        
    class Test1(Parameterized):
        parameters_default = [param]

        
    t = Test1()
    assert t.parameters['a'] == 1
    assert t.get_parameter_default('a') == 1

    t = Test1(parameters={'a': 2})
    assert t.parameters['a'] == 2
    assert t.get_parameter_default('a') == 1
    
    with pytest.raises(ValueError):
        t = Test1(parameters={'b': 3})
    t = Test1()
    ps = t._parse_parameters({'b': 3}, check_validity=False)
    assert ps['a'] == 1
    assert ps['b'] == 3
    
    
    class Test2(Test1):
        # also test conversion of default parameters
        parameters_default = [Parameter('b', '2', int, "help")]
        
    t = Test2()
    assert t.parameters['a'] == 1
    assert t.parameters['b'] == 2
    
    t = Test2(parameters={'a': 10, 'b': 20})
    assert t.parameters['a'] == 10
    assert t.parameters['b'] == 20
    assert t.get_parameter_default('a') == 1
    assert t.get_parameter_default('b') == '2'
    with pytest.raises(KeyError):
        t.get_parameter_default('c')
        
    class Test3(Test2):
        # test overwriting defaults
        parameters_default = [Parameter('a', 3), Parameter('c', 4)]
        
    t = Test3()
    assert t.parameters['a'] == 3
    assert t.get_parameter_default('a') == 3
    assert set(t.parameters.keys()) == {'a', 'b', 'c'}
        
        
        
def test_parameters_simple():
    """ test adding parameters using a simple dictionary """
    
    class Test(Parameterized):
        parameters_default = {'a': 1}
        
    t = Test()
    assert t.parameters['a'] == 1
        
        
        
def test_parameter_help(capsys):
    """ test how parameters are shown """
    class Test1(Parameterized):
        parameters_default = [DeprecatedParameter('a', 1, int, "random string")]
        
    class Test2(Test1):
        parameters_default = [Parameter('b', 2, int, "another word")]
        
    t = Test2()
    for flags in itertools.combinations_with_replacement([True, False], 3):
        Test2.show_parameters(*flags)
        o1, e1 = capsys.readouterr()
        t.show_parameters(*flags)
        o2, e2 = capsys.readouterr()
        assert o1 == o2
        assert e1 == e2 == ''
    
    
    
def test_hidden_parameter():
    """ test how hidden parameters are handled """
    class Test1(Parameterized):
        parameters_default = [Parameter('a', 1), Parameter('b', 2)]
         
    assert Test1().parameters == {'a': 1, 'b': 2}
    
    class Test2(Test1):
        parameters_default = [HideParameter('b')]
        
    class Test2a(Parameterized):
        parameters_default = [Parameter('a', 1), Parameter('b', 2, hidden=True)]
        
    for t_class in [Test2, Test2a]:
        assert 'b' not in t_class.get_parameters()
        assert len(t_class.get_parameters()) == 1
        assert len(t_class.get_parameters(include_hidden=True)) == 2
        t2 = t_class()
        assert t2.parameters == {'a': 1, 'b': 2}
        assert t2.get_parameter_default('b') == 2
        with pytest.raises(ValueError):
            t2._parse_parameters({'b': 2}, check_validity=True,
                                 allow_hidden=False)
        
    class Test3(Test1):
        parameters_default = [Parameter('b', 3)]
    
    t3 = Test3()
    assert t3.parameters == {'a': 1, 'b': 3}
    assert t3.get_parameter_default('b') == 3
    
    
    
def test_get_all_parameters():
    """ test the get_all_parameters function """
    p1 = get_all_parameters()
    for key in ['value', 'description']:
        p2 = get_all_parameters(key)
        assert set(p1) == p2.keys()
    
    


"""
Base classes
   
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de> 
"""

from abc import ABCMeta, abstractmethod
import logging
from typing import (Callable, Optional, TYPE_CHECKING, Union,  # @UnusedImport
                    Dict, Tuple, Any)

import numpy as np

from ..fields import FieldCollection
from ..fields.base import FieldBase, OptionalArrayLike
from ..trackers.base import TrackerCollectionDataType
from ..tools.numba import jit


if TYPE_CHECKING:
    from ..solvers.controller import TRangeType  # @UnusedImport



class PDEBase(metaclass=ABCMeta):
    """ base class for solving partial differential equations """

    check_implementation: bool = True
    """ bool: Flag determining whether (some) numba-compiled functions should be
    checked against their numpy counter-parts. This can help with implementing a
    correct compiled version for a PDE class. """
    
    explicit_time_dependence: Optional[bool] = None
    """ bool: Flag indicating whether the right hand side of the PDE has an
    explicit time dependence. """


    def __init__(self, noise: OptionalArrayLike = 0):
        """
        Args:
            noise (float or :class:`numpy.ndarray`):
                Magnitude of the additive Gaussian white noise that is supported
                by default. If set to zero, a deterministic partial differential
                equation will be solved. Different noise magnitudes can be
                supplied for each field in coupled PDEs.
                
        Note:
            If more complicated noise structures are required, the methods
            :meth:`PDEBase.noise_realization` and
            :meth:`PDEBase._make_noise_realization_numba` need to be overwritten
            for the `numpy` and `numba` backend, respectively.
        """
        self._logger = logging.getLogger(self.__class__.__name__)
        self.noise = noise


    @property
    def is_sde(self) -> bool:
        """ flag indicating whether this is a stochastic differential equation
        
        The :class:`BasePDF` class supports additive Gaussian white noise, whose
        magnitude is controlled by the `noise` property. In this case, `is_sde`
        is `True` if `self.noise != 0`.
        """
        # check for self.noise, in case __init__ is not called in a subclass
        return hasattr(self, 'noise') and self.noise != 0


    @abstractmethod
    def evolution_rate(self, field: FieldBase, t: float = 0) \
        -> FieldBase: pass


    def _make_pde_rhs_numba(self, state: FieldBase) -> Callable:
        """ create a compiled function for evaluating the right hand side """
        raise NotImplementedError


    def make_pde_rhs(self, state: FieldBase, backend: str = 'auto') -> Callable:
        """ return a function for evaluating the right hand side of the PDE
        
        Args:
            state (:class:`~pde.fields.FieldBase`):
                An example for the state from which the grid and other
                information can be extracted
            backend (str): Determines how the function is created. Accepted 
                values are 'python` and 'numba'. Alternatively, 'auto' lets the
                code decide for the most optimal backend.
                
        Returns:
            Function determining the right hand side of the PDE
        """
        if backend == 'auto':
            try:
                rhs = self._make_pde_rhs_numba(state)
            except NotImplementedError:
                backend = 'numpy'
            else:
                rhs._backend = 'numba'  # type: ignore
            
        if backend == 'numba':
            rhs = self._make_pde_rhs_numba(state)
            rhs._backend = 'numba'  # type: ignore
                
        elif backend == 'numpy':
            state = state.copy()
            
            def evolution_rate_numpy(state_data, t: float):
                """ evaluate the rhs given only a state without the grid """
                state.data = state_data
                return self.evolution_rate(state, t).data
        
            rhs = evolution_rate_numpy
            rhs._backend = 'numpy'  # type: ignore
            
        elif backend != 'auto':
            raise ValueError(f"Unknown backend `{backend}`. Possible values "
                             "are ['auto', 'numpy', 'numba']")
        
        if (self.check_implementation and
                rhs._backend == 'numba'):  # type: ignore
            # compare the numba implementation to the numpy implementation
            expected = self.evolution_rate(state.copy()).data
            test_state = state.copy()
            result = rhs(test_state.data, 0)
            if not np.allclose(result, expected):
                raise RuntimeError('The numba compiled implementation of the '
                                   'right hand side is not compatible with '
                                   'the numpy implementation. This check can '
                                   'be disabled by setting the class attribute '
                                   '`check_implementation` to `False`.')
        
        return rhs
            
            
    def noise_realization(self, state: FieldBase, t: float = 0,
                          label: str = 'Noise realization') -> FieldBase:
        """ returns a realization for the noise
        
        Args:
            state (:class:`~pde.fields.ScalarField`):
                The scalar field describing the concentration distribution
            t (float):
                The current time point
            label (str):
                The label for the returned field
            
        Returns:
            :class:`~pde.fields.ScalarField`:
            Scalar field describing the evolution rate of the PDE 
        """
        if self.noise:
            if np.isscalar(self.noise):
                # a single noise value is given for all fields
                data = np.random.normal(scale=self.noise, size=state.data.shape)
                return state.copy(data=data, label=label)
            
            elif isinstance(state, FieldCollection):
                # different noise strengths, assuming one for each field
                noise = np.broadcast_to(self.noise, len(state))
                fields = [f.copy(data=np.random.normal(scale=n,
                                                       size=f.data.shape))
                          for f, n in zip(state, noise)]
                return FieldCollection(fields, label=label)
            
            else:
                # different noise strengths, but a single field 
                raise RuntimeError('Multiple noise strengths were given for '
                                   f'the single field {state}')
                
        else:
            return state.copy(data=0, label=label)

       
    def _make_noise_realization_numba(self, state: FieldBase) -> Callable:            
        """ return a function for evaluating the noise term of the PDE
        
        Args:
            state (:class:`~pde.fields.FieldBase`):
                An example for the state from which the grid and other
                information can be extracted
                
        Returns:
            Function determining the right hand side of the PDE
        """
        if self.noise:        
            data_shape = state.data.shape
            
            if np.isscalar(self.noise):
                # a single noise value is given for all fields
                noise_strength = float(self.noise)
                            
                @jit
                def noise_realization(state_data: np.ndarray, t: float):
                    """ helper function returning a noise realization """ 
                    return noise_strength * np.random.randn(*data_shape)
                
            elif isinstance(state, FieldCollection):
                # different noise strengths, assuming one for each field
                noise_strengths = np.empty(data_shape[0])
                noise_arr = np.broadcast_to(self.noise, len(state))
                for i, noise in enumerate(noise_arr):
                    noise_strengths[state._slices[i]] = noise
                
                @jit
                def noise_realization(state_data: np.ndarray, t: float):
                    """ helper function returning a noise realization """ 
                    out = np.random.randn(*data_shape)
                    for i in range(data_shape[0]):
                        out[i] *= noise_strengths[i]
                    return out
                
            else:
                # different noise strengths, but a single field 
                raise RuntimeError('Multiple noise strengths were given for '
                                   f'the single field {state}')
            
        else:
            @jit
            def noise_realization(state_data: np.ndarray, t: float):
                """ helper function returning a noise realization """ 
                return None
        
        return noise_realization  # type: ignore    
            
       
    def _make_sde_rhs_numba(self, state: FieldBase) -> Callable:            
        """ return a function for evaluating the noise term of the PDE
        
        Args:
            state (:class:`~pde.fields.FieldBase`):
                An example for the state from which the grid and other
                information can be extracted
                
        Returns:
            Function determining the right hand side of the PDE
        """
        evolution_rate = self._make_pde_rhs_numba(state)
        noise_realization = self._make_noise_realization_numba(state)
        
        @jit
        def sde_rhs(state_data: np.ndarray, t: float):
            """ compiled helper function returning a noise realization """ 
            return (evolution_rate(state_data, t),
                    noise_realization(state_data, t))
        
        return sde_rhs  # type: ignore    
    
                        
    def make_sde_rhs(self, state: FieldBase, backend: str = 'auto') \
            -> Callable:
        """ return a function for evaluating the right hand side of the SDE
        
        Args:
            state (:class:`~pde.fields.FieldBase`):
                An example for the state from which the grid and other
                information can be extracted
            backend (str): Determines how the function is created. Accepted 
                values are 'python` and 'numba'. Alternatively, 'auto' lets the
                code decide for the most optimal backend.
                
        Returns:
            Function determining the deterministic part of the right hand side
            of the PDE together with a noise realization.
        """
        if backend == 'auto':
            try:
                sde_rhs = self._make_sde_rhs_numba(state)
            except NotImplementedError:
                backend = 'numpy'
            else:
                sde_rhs._backend = 'numba'  # type: ignore
                return sde_rhs
             
        if backend == 'numba':
            sde_rhs = self._make_sde_rhs_numba(state)
            sde_rhs._backend = 'numba'  # type: ignore
                
        elif backend == 'numpy':
            state = state.copy()
            
            def sde_rhs(state_data, t: float):
                """ evaluate the rhs given only a state without the grid """
                state.data = state_data
                return (self.evolution_rate(state, t).data,
                        self.noise_realization(state, t).data)
        
            sde_rhs._backend = 'numpy'  # type: ignore
            
        else:
            raise ValueError(f'Unknown backend `{backend}`')
        
        return sde_rhs
            

    def solve(self, state: FieldBase,
              t_range: "TRangeType",
              dt: float = None,
              tracker: TrackerCollectionDataType = ['progress', 'consistency'],
              method: str = 'auto',
              ret_info: bool = False,
              **kwargs) -> Union[FieldBase, Tuple[FieldBase, Dict[str, Any]]]:
        """ convenience method for solving the partial differential equation 
        
        The method constructs a suitable solver
        (:class:`~pde.solvers.base.SolverBase`) and controller
        (:class:`~pde.controller.Controller`) to advance the state over the
        temporal range specified by `t_range`. To obtain full flexibility, it is
        advisable to construct these classes explicitly. 

        Args:
            state (:class:`~pde.fields.base.FieldBase`):
                The initial state (which also defines the grid)
            t_range (float or tuple):
                Sets the time range for which the PDE is solved. If only a
                single value `t_end` is given, the time range is assumed to be 
                `[0, t_end]`.
            dt (float):
                Time step of the chosen stepping scheme. If `None`, a default
                value based on the stepper will be chosen.
            tracker:
                Defines a tracker that process the state of the simulation at
                fixed time intervals. Multiple trackers can be specified as a
                list. The default value is ['progress', 'consistency'], which
                displays a progress bar and checks the state for consistency,
                aborting the simulation when not-a-number values appear.
            method (:class:`~pde.solvers.base.SolverBase` or str):
                Specifies a method for solving the differential equation. This
                can either be an instance of
                :class:`~pde.solvers.base.SolverBase` or a descriptive name
                like 'explicit' or 'scipy'. The valid names are given by
                :meth:`pde.solvers.base.SolverBase.registered_solvers`.
            ret_info (bool):
                Flag determining whether diagnostic information about the solver
                process should be returned.
            **kwargs:
                Additional keyword arguments are forwarded to the solver class
                
        Returns:
            :class:`~pde.fields.base.FieldBase`:
            The state at the final time point. In the case `ret_info == True`, a
            tuple with the final state and a dictionary with additional
            information is returned.
        """
        from ..solvers.base import SolverBase
        
        if method == 'auto':
            method = 'scipy' if dt is None else 'explicit'
        
        # create solver
        if callable(method):
            solver = method(pde=self, **kwargs)
            if not isinstance(solver, SolverBase):
                self._logger.warn('Solver is not an instance of `SolverBase`. '
                                  'Specified wrong method?')
        else:
            solver = SolverBase.from_name(method, pde=self, **kwargs)
        
        # create controller
        from ..solvers import Controller
        controller = Controller(solver, t_range=t_range, tracker=tracker)
        
        # run the simulation
        final_state = controller.run(state, dt)
        
        if ret_info:
            info = controller.info.copy()
            info.pop('solver_class')  # remove redundant information
            info['solver'] = solver.info.copy()
            return final_state, info
        else: 
            return final_state
                


"""
Solver comparison
=================

This example shows how to set up solvers explicitly and how to extract
diagnostic information.
"""

from pde import (UnitGrid, ScalarField, FieldCollection, DiffusionPDE,
                 ExplicitSolver, ScipySolver, Controller) 

# initialize the grid, an initial condition, and the PDE
grid = UnitGrid([32, 32])
field = ScalarField.random_uniform(grid, -1, 1)
eq = DiffusionPDE()

# try the explicit solver
solver1 = ExplicitSolver(eq)
controller1 = Controller(solver1, t_range=1, tracker=None)
sol1 = controller1.run(field, dt=1e-3)
sol1.label = 'explicit solver'
print('Diagnostic information from first run:')
print(controller1.diagnostics)
print()

# try the standard scipy solver
solver2 = ScipySolver(eq)
controller2 = Controller(solver2, t_range=1, tracker=None)
sol2 = controller2.run(field)
sol2.label = 'scipy solver'
print('Diagnostic information from second run:')
print(controller2.diagnostics)
print()

# plot both fields and give the deviation as the title
title = f'Deviation: {((sol1 - sol2)**2).average:.2g}'
FieldCollection([sol1, sol2]).plot(title=title)



'''
This module defines the :class:`Cuboid` class, which represents an n-dimensional
cuboid that is aligned with the axes of a Cartesian coordinate system.

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import itertools
from typing import Tuple

import numpy as np



class Cuboid:
    """ class that represents a cuboid in :math:`n` dimensions """
    
    
    def __init__(self, pos, size, mutable: bool = True):
        """ defines a cuboid from a position and a size vector 

        Args:
            pos (list):
                The position of the lower left corner. The length of this list
                determines the dimensionality of space
            size (list):
                The size of the cuboid along each dimension.
            mutable (bool):
                Flag determining whether the cuboid parameters can be changed
        """
        self._mutable = mutable
        # set position and adjust mutable status later
        self.pos = np.array(pos, copy=True)
        self.size = size  # implicitly sets correct shape
        self.pos.flags.writeable = self.mutable 
        
        
    @property
    def size(self):
        return self._size
    
    @size.setter
    def size(self, value):
        self._size = np.array(value, self.pos.dtype)  # make copy
        if self.pos.shape != self._size.shape:
            raise ValueError('Size vector (dim=%d) must have the same '
                             'dimension as the position vector (dim=%d)' %
                             (len(self._size), len(self.pos)))
        
        # flip Cuboid with negative size
        neg = (self._size < 0)
        self.pos[neg] += self._size[neg]
        self._size = np.abs(self._size)
        self._size.flags.writeable = self.mutable
        

    @property
    def corners(self):
        """ return coordinates of two extreme corners defining the cuboid """
        return np.copy(self.pos), self.pos + self.size
    
            
    @property
    def mutable(self) -> bool:
        return self._mutable
    
    @mutable.setter
    def mutable(self, value: bool):
        self._mutable = bool(value)
        self.pos.flags.writeable = self._mutable
        self._size.flags.writeable = self._mutable

        
    @classmethod
    def from_points(cls, p1, p2, **kwargs) -> "Cuboid":
        """ create cuboid from two points
        
        Args:
            p1 (list): Coordinates of first corner point
            p2 (list): Coordinates of second corner point
            
        Returns:
            Cuboid: cuboid with positive size
        """
        p1 = np.asarray(p1)
        p2 = np.asarray(p2)
        return cls(p1, p2 - p1, **kwargs)
    
    
    @classmethod
    def from_bounds(cls, bounds, **kwargs) -> "Cuboid":
        """ create cuboid from bounds
        
        Args:
            bounds (list): Two dimensional array of axes bounds
            
        Returns:
            Cuboid: cuboid with positive size
        """
        bounds = np.asarray(bounds).reshape(-1, 2)
        return cls(bounds[:, 0], bounds[:, 1] - bounds[:, 0], **kwargs)
    
    
    @classmethod
    def from_centerpoint(cls, centerpoint, size, **kwargs) \
            -> "Cuboid":
        """ create cuboid from two points
        
        Args:
            centerpoint (list): Coordinates of the center
            size (list): Size of the cuboid
            
        Returns:
            Cuboid: cuboid with positive size
        """        
        centerpoint = np.asarray(centerpoint)
        size = np.asarray(size)
        return cls(centerpoint - size / 2, size, **kwargs)
    
    
    def copy(self) -> "Cuboid":
        return self.__class__(self.pos, self.size)
        
        
    def __repr__(self):
        return "{cls}(pos={pos}, size={size})".format(
                        cls=self.__class__.__name__,
                        pos=self.pos, size=self.size)
            
            
    def __add__(self, other: "Cuboid"):
        """ The sum of two cuboids is the minimal cuboid enclosing both """
        if isinstance(other, Cuboid):
            if self.dim != other.dim:
                raise RuntimeError('Incompatible dimensions')
            a1, a2 = self.corners
            b1, b2 = other.corners
            return self.__class__.from_points(np.minimum(a1, b1),
                                              np.maximum(a2, b2))
            
        else:
            raise NotImplementedError('Cannot add cuboid and '
                                      f'{other.__class__.__name__}')
            
            
    def __eq__(self, other):
        """ override the default equality test """
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (np.all(self.pos == other.pos) and 
                np.all(self.size == other.size))

    
    @property
    def dim(self) -> int:
        return len(self.pos)
            

    @property
    def bounds(self) -> Tuple[Tuple[float, float], ...]:
        return tuple((p, p + s) for p, s in zip(self.pos, self.size))
        

    @property
    def vertices(self):
        """ return the coordinates of all the corners """
        return list(itertools.product(*self.bounds))


    @property
    def diagonal(self) -> float:
        """ returns the length of the diagonal """
        return np.linalg.norm(self.size)  # type: ignore
    
    
    @property
    def surface_area(self) -> float:
        """ surface area of a cuboid in :math:`n` dimensions.
        
        The surface area is the volume of the (:math:`n-1`)-dimensional
        hypercubes that bound the current cuboid:
        
            * :math:`n=1`: the number of end points (2)
            * :math:`n=2`: the perimeter of the rectangle
            * :math:`n=3`: the surface area of the cuboid
        """
        sides = self.size
        null = (sides == 0)
        null_count = null.sum()
        if null_count == 0:
            return 2 * np.sum(np.product(sides) / sides)  # type: ignore
        elif null_count == 1:
            return 2 * np.product(sides[~null])  # type: ignore
        else:
            return 0


    @property
    def centroid(self):
        return self.pos + self.size / 2
    
    @centroid.setter
    def centroid(self, center):
        self.pos[:] = np.asanyarray(center) - self.size / 2
    
    
    @property
    def volume(self) -> float:
        return np.prod(self.size)  # type: ignore
    

    def buffer(self, amount=0, inplace=False):
        """ dilate the cuboid by a certain amount in all directions """
        amount = np.asarray(amount)
        if inplace:
            self.pos -= amount
            self.size += 2*amount
            return self
        else:
            return self.__class__(self.pos - amount, self.size + 2*amount)
    
        
    def contains_point(self, points):
        """ returns a True when `points` are within the Cuboid
        
        Args:
            points (list): List of point coordinates
            
        Returns:
            np.ndarray: list of booleans indicating which points are inside
        """
        points = np.asarray(points)
        if len(points) == 0:
            return points
        
        if points.shape[-1] != self.dim:
            raise ValueError('Last dimension of `points` must agree with '
                             'cuboid dimension %d' % self.dim)
        
        c1, c2 = self.corners
        return np.all(c1 <= points, axis=-1) & np.all(points <= c2, axis=-1)
        


def asanyarray_flags(data: np.ndarray, dtype=None, writeable: bool = True):
    """ turns data into an array and sets the respective flags.
    
    A copy is only made if necessary
    
    Args:
        data (:class:`numpy.ndarray`): numpy array that whose flags are adjusted
        dtype: the resulant dtype
        writeable (bool): Flag determining whether the results is writable
        
    Returns:
        :class:`numpy.ndarray`:
            array with same data as `data` but with flags adjusted.
    """
    try:
        data_writeable = data.flags.writeable
    except AttributeError:
        # `data` did not have the writeable flag => it's not a numpy array  
        result = np.array(data, dtype)
    else:
        if data_writeable != writeable:
            # need to make a copy because the flags differ
            result = np.array(data, dtype)
        else:
            # might have to make a copy to adjust the dtype
            result = np.asanyarray(data, dtype)
            
    # set the flags and return the array
    result.flags.writeable = writeable
    return result
    


'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import numpy as np

import pytest

from ..vectorial import VectorField
from ..base import FieldBase
from ...grids import UnitGrid, CartesianGrid
from ...tools.misc import module_available



def test_vectors():
    """ test some vector fields """
    grid = CartesianGrid([[0.1, 0.3], [-2, 3]], [3, 4])
    v1 = VectorField(grid, np.full((2,) + grid.shape, 1))
    v2 = VectorField(grid, np.full((2,) + grid.shape, 2))
    np.testing.assert_allclose(v1.average, (1, 1))
    assert np.allclose(v1.magnitude, np.sqrt(2))

    v3 = v1 + v2
    assert v3.grid == grid
    np.testing.assert_allclose(v3.data, 3)
    v1 += v2
    np.testing.assert_allclose(v1.data, 3)
    
    # test projections
    v1 = VectorField(grid)
    v1.data[0, :] = 3
    v1.data[1, :] = 4
    for method, value in [('min', 3), ('max', 4), ('norm', 5),
                          ('squared_sum', 25), ('auto', 5)]:
        p1 = v1.to_scalar(method)
        assert p1.data.shape == grid.shape
        np.testing.assert_allclose(p1.data, value)

    v2 = FieldBase.from_state(v1.attributes, data=v1.data)
    assert v1 == v2
    assert v1.grid is v2.grid
    
    attrs = VectorField.unserialize_attributes(v1.attributes_serialized)
    v2 = FieldBase.from_state(attrs, data=v1.data)
    assert v1 == v2
    assert v1.grid is not v2.grid

    # test dot product
    v2._grid = v1.grid  # make sure grids are identical
    v1.data = 1
    v2.data = 2
    for s in (v1 @ v2, v2 @ v1, v1.dot(v2)):
        from ..scalar import ScalarField
        assert isinstance(s, ScalarField)
        assert s.grid is grid
        np.testing.assert_allclose(s.data, np.full(grid.shape, 4))
        
    # test options for plotting images
    if module_available("matplotlib"):
        v1.plot(method='streamplot', transpose=True)


    
def test_divergence():
    """ test the divergence operator """
    grid = CartesianGrid([[0, 2 * np.pi], [0, 2 * np.pi]], [16, 16],
                         periodic=True)
    x, y = grid.cell_coords[..., 0], grid.cell_coords[..., 1]
    data = [np.cos(x) + y, np.sin(y) - x]
    v = VectorField(grid, data)
    s = v.divergence('natural')
    assert s.data.shape == (16, 16)
    div = np.cos(y) - np.sin(x)
    np.testing.assert_allclose(s.data, div, rtol=0.1, atol=0.1)


    
def test_vector_laplace():
    """ test the laplace operator """
    grid = CartesianGrid([[0, 2 * np.pi], [0, 2 * np.pi]], [16, 16],
                         periodic=True)
    x, y = grid.cell_coords[..., 0], grid.cell_coords[..., 1]
    data = [np.cos(x) + np.sin(y), np.sin(y) - np.cos(x)]
    v = VectorField(grid, data)
    vl = v.laplace('natural')
    assert vl.data.shape == (2, 16, 16)
    np.testing.assert_allclose(vl.data[0, ...], -np.cos(x) - np.sin(y),
                               rtol=0.1, atol=0.1)
    np.testing.assert_allclose(vl.data[1, ...], -np.sin(y) + np.cos(x),
                               rtol=0.1, atol=0.1)



def test_vector_boundary_conditions():
    """ test some boundary conditions of operators of vector fields """
    grid = CartesianGrid([[0, 2*np.pi], [0, 1]], 32, periodic=False)
    v_x = np.sin(grid.cell_coords[..., 0])
    v_y = grid.cell_coords[..., 1]
    vf = VectorField(grid, np.array([v_x, v_y]))
    
    bc_x = [{'type': 'derivative', 'value': [0, -1]},
            {'type': 'derivative', 'value': [0, 1]}]
    bc_y = [{'type': 'value', 'value': [0, 0]},
            {'type': 'value', 'value': [1, 1]}]
    tf = vf.gradient(bc=[bc_x, bc_y])
    
    np.testing.assert_allclose(tf[0, 1].data[1:-1, :], 0)
    np.testing.assert_allclose(tf[1, 1].data, 1)



def test_outer_product():
    """ test outer product of vector fields """
    vf = VectorField(UnitGrid([1, 1]), [[[1]], [[2]]])
    tf = vf.outer_product(vf)
    np.testing.assert_equal(tf.data, np.array([1, 2, 2, 4]).reshape(2, 2, 1, 1))
    tf.data = 0
    vf.outer_product(vf, out=tf)
    np.testing.assert_equal(tf.data, np.array([1, 2, 2, 4]).reshape(2, 2, 1, 1))    
    
    
    
def test_from_expressions():
    """ test initializing vector fields with expressions """
    grid = UnitGrid([4, 4])
    vf = VectorField.from_expression(grid, ['x**2', 'x * y'])
    xs = grid.cell_coords[..., 0]
    ys = grid.cell_coords[..., 1]
    np.testing.assert_allclose(vf.data[0], xs**2)
    np.testing.assert_allclose(vf.data[1], xs * ys)

    # corner case
    vf = VectorField.from_expression(grid, ['1', 'x * y'])
    
    with pytest.raises(ValueError):
        VectorField.from_expression(grid, 'xy')
    with pytest.raises(ValueError):
        VectorField.from_expression(grid, ['xy'])
    with pytest.raises(ValueError):
        VectorField.from_expression(grid, ['x'] * 3)
        
        

def test_vector_plot_quiver_reduction():
    """ test whether quiver plots reduce the resolution """
    grid = UnitGrid([6, 6])
    field = VectorField.random_normal(grid)
    ref = field.plot(method='quiver', max_points=4)
    assert len(ref.element.U) == 16
        


def test_boundary_interpolation_vector():
    """ test boundary interpolation """
    grid = CartesianGrid([[0.1, 0.3], [-2, 3]], [3, 3])
    field = VectorField.random_normal(grid)
    
    # test boundary interpolation
    bndry_val = np.random.randn(2, 3)
    for bndry in grid._iter_boundaries():
        val = field.get_boundary_values(*bndry, bc={'value': bndry_val})
        np.testing.assert_allclose(val, bndry_val)
    
        ev = field.make_get_boundary_values(*bndry, bc={'value': bndry_val})
        np.testing.assert_allclose(ev(), bndry_val)
    
            


'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import itertools

import numpy as np

from ... import UnitGrid
from ..axis import get_boundary_axis
from ..axes import Boundaries
from ....fields import ScalarField

    
 
def test_boundaries():
    """ test setting boundaries for multiple systems """
    b = ['periodic', 'value', {'type': 'derivative', 'value': 1}]
    for bx, by in itertools.product(b, b):
        g = UnitGrid([2, 2], periodic=[b == 'periodic' for b in (bx, by)])

        bcs = Boundaries.from_data(g, [bx, by])
        bc_x = get_boundary_axis(g, 0, bx)
        bc_y = get_boundary_axis(g, 1, by)
        
        assert bcs.grid.num_axes == 2
        assert bcs[0] == bc_x
        assert bcs[1] == bc_y
        assert bcs == Boundaries.from_data(g, [bc_x, bc_y])
        if bx == by:
            assert bcs == Boundaries.from_data(g, bx)
            
        bc2 = bcs.copy()
        assert bcs == bc2
        assert bcs is not bc2
        
    b1 = Boundaries.from_data(UnitGrid([2, 2]), 'natural')
    b2 = Boundaries.from_data(UnitGrid([3, 3]), 'natural')
    assert b1 != b2
        
        
        
def test_boundary_specifications():
    """ test different ways of specifying boundary conditions """
    g = UnitGrid([2])
    bc1 = Boundaries.from_data(g, [{'type': 'derivative', 'value': 0},
                                   {'type': 'value', 'value': 0}])
    assert bc1 == Boundaries.from_data(g, [{'type': 'derivative'},
                                           {'type': 'value'}])
    assert bc1 == Boundaries.from_data(g, [{'derivative': 0},
                                           {'value': 0}])
    assert bc1 == Boundaries.from_data(g, ['no-flux', 'value'])
    assert bc1 == Boundaries.from_data(g, ['neumann', 'dirichlet'])
    
    

def test_mixed_boundary_condition():
    """ test limiting cases of the mixed boundary condition """
    g = UnitGrid([2])
    d = np.random.random(2)
    g1 = g.get_operator('gradient', bc=[{'mixed': 0}, {'mixed': np.inf}])
    g2 = g.get_operator('gradient', bc=['derivative', 'value'])
    np.testing.assert_allclose(g1(d), g2(d))
        
        
        
def test_special_cases():
    """ test some special boundary conditions """
    g = UnitGrid([5])
    s = ScalarField(g, np.arange(5))
    for bc in ['extrapolate', {'curvature': 0}]:
        np.testing.assert_allclose(s.laplace(bc).data, 0)
    


'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import itertools
import pytest

from ... import UnitGrid
from ..local import BCBase
from ..axis import BoundaryPair, get_boundary_axis



def test_boundary_pair():
    """ test setting boundary conditions for whole axis """
    g = UnitGrid([2, 3])
    b = ['value', {'type': 'derivative', 'value': 1}]
    for bl, bh in itertools.product(b, b):
        bc = BoundaryPair.from_data(g, 0, [bl, bh])
        blo = BCBase.from_data(g, 0, upper=False, data=bl)
        bho = BCBase.from_data(g, 0, upper=True, data=bh)
        
        assert bc.low == blo
        assert bc.high == bho
        assert bc == BoundaryPair(blo, bho)
        if bl == bh:
            assert bc == BoundaryPair.from_data(g, 0, bl)
        assert list(bc) == [blo, bho]
                      
        bc.check_value_rank(0)
        with pytest.raises(RuntimeError):
            bc.check_value_rank(1)
                
    data = {'low': {'value': 1}, 'high': {'derivative': 2}}
    bc1 = BoundaryPair.from_data(g, 0, data)
    bc2 = BoundaryPair.from_data(g, 0, data)
    assert bc1 == bc2 and bc1 is not bc2
    bc2 = BoundaryPair.from_data(g, 1, data)
    assert bc1 != bc2 and bc1 is not bc2
    
    # miscellaneous methods
    bc1.set_value(0)
    b_lo, b_hi = bc1
    assert b_lo == BCBase.from_data(g, 0, False, {'value': 0})
    assert b_hi == BCBase.from_data(g, 0, True, {'derivative': 0})
    
                            
                            
def test_get_axis_boundaries():
    """ test setting boundary conditions including periodic ones """
    g = UnitGrid([2])
    for data in ['value', 'derivative', 'periodic']:
        b = get_boundary_axis(g, 0, data)
        assert str(b) == '"' + data + '"'
        
        if data == 'periodic':
            assert b.periodic
            assert len(list(b)) == 0
        else:
            assert not b.periodic
            assert len(list(b)) == 2
            


'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import pytest
import numpy as np

from ..expressions import parse_number, ScalarExpression, TensorExpression



def test_parse_number():
    """ test parse_number function """
    assert parse_number(0) == pytest.approx(0)
    assert parse_number(1.235) == pytest.approx(1.235)
    assert parse_number("0") == pytest.approx(0)
    assert parse_number("1.235") == pytest.approx(1.235)
    assert parse_number("-1.4e-5") == pytest.approx(-1.4e-5)
    assert parse_number("sin(3.4)") == pytest.approx(np.sin(3.4))
    assert parse_number("sin(a)", {'a': 3}) == pytest.approx(np.sin(3))
    assert parse_number("a**b", {'a': 3.2, 'b': 4.5}) == pytest.approx(3.2**4.5)
    with pytest.raises(TypeError):
        parse_number('foo')



def test_const():
    """ test simple expressions """
    for expr in [None, 1, '1', 'a - a']:
        e = ScalarExpression() if expr is None else ScalarExpression(expr)
        val = 0 if expr is None or expr == 'a - a' else float(expr)
        assert e.constant
        assert e.value == val
        assert e() == val
        assert e.get_compiled()() == val
        assert not e.depends_on('a')
        assert e.differentiate('a').value == 0
        assert e.shape == tuple()
        assert e.rank == 0
        assert bool(e) == (val != 0)
        assert e.is_zero == (val == 0)
        
        g = e.derivatives
        assert g.constant
        np.testing.assert_equal(g.value, [])
            
        for f in [ScalarExpression(e), ScalarExpression(e.expression),
                  ScalarExpression(e.value)]:
            assert e is not f
            assert e._sympy_expr == f._sympy_expr



def test_single_arg():
    """ test simple expressions """
    e = ScalarExpression('2 * a')
    assert not e.constant
    assert e.depends_on('a')
    assert e(4) == 8
    assert e.get_compiled()(4) == 8
    assert e.differentiate('a').value == 2
    assert e.differentiate('b').value == 0
    assert e.shape == tuple()
    assert e.rank == 0
    assert bool(e)
    assert not e.is_zero
    
    assert e == ScalarExpression(e.expression)
    with pytest.raises(TypeError):
        e.value
    
    arr = np.random.random(5)
    np.testing.assert_allclose(e(arr), 2 * arr)
    np.testing.assert_allclose(e.get_compiled()(arr), 2 * arr)
    
    g = e.derivatives
    assert g.shape == (1,)
    assert g.constant
    assert g(3) == [2]
    assert g.get_compiled()(3) == [2]
    
    with pytest.raises(TypeError):
        ScalarExpression(np.exp)



def test_two_args():
    """ test simple expressions """
    e = ScalarExpression('2 * a ** b')
    assert e.depends_on('b')
    assert not e.constant
    assert e(4, 2) == 32
    assert e.get_compiled()(4, 2) == 32
    assert e.differentiate('a')(4, 2) == 16
    assert e.differentiate('b')(4, 2) == pytest.approx(32*np.log(4))
    assert e.differentiate('c').value == 0
    assert e.shape == tuple()
    assert e.rank == 0
    assert e == ScalarExpression(e.expression)

    for x in [np.random.random(2), np.random.random((2, 5))]:
        res = 2 * x[0]**x[1]
        np.testing.assert_allclose(e(*x), res)
        np.testing.assert_allclose(e.get_compiled()(*x), res)
        if x.ndim == 1:
            func = e._get_function(single_arg=True)
            np.testing.assert_allclose(func(x), res)
            func = e.get_compiled(single_arg=True)
            np.testing.assert_allclose(func(x), res)

    g = e.derivatives
    assert g.shape == (2,)
    assert g.rank == 1
    assert not g.constant
    np.testing.assert_allclose(g(2, 3), [24, 16 * np.log(2)])
    np.testing.assert_allclose(g.get_compiled()(2, 3), [24, 16 * np.log(2)])
    
    
    
def test_derivatives():
    """ test vector expressions """
    e = ScalarExpression('a * b**2')
    assert e.depends_on('a') and e.depends_on('b')
    assert not e.constant
    assert e.rank == 0
    
    d1 = e.derivatives
    assert d1.shape == (2,)
    np.testing.assert_allclose(d1(2, 3), [9, 12])
    np.testing.assert_allclose(d1.get_compiled()(2, 3), [9, 12])
    
    d2 = d1.derivatives
    assert d2.shape == (2, 2)
    np.testing.assert_allclose(d2(2, 3), [[0, 6], [6, 4]])
    np.testing.assert_allclose(d2.get_compiled()(2, 3), [[0, 6], [6, 4]])
    
    d3 = d2.derivatives
    assert d3.shape == (2, 2, 2)

    d4 = d3.derivatives
    assert d4.shape == (2, 2, 2, 2)
    np.testing.assert_allclose(d4(2, 3), np.zeros((2, 2, 2, 2)))
    np.testing.assert_allclose(d4.get_compiled()(2, 3),
                               np.zeros((2, 2, 2, 2)))

            
            
def test_indexed():
    """ test simple expressions """
    e = ScalarExpression('2 * a[0] ** a[1]', allow_indexed=True)
    assert not e.constant
    assert e.depends_on('a')
    
    a = np.array([4, 2])
    assert e(a) == 32
    assert e.get_compiled()(a) == 32
    
    assert e.differentiate('a[0]')(a) == 16
    assert e.differentiate('a[1]')(a) == pytest.approx(32*np.log(4))
    
    with pytest.raises(RuntimeError):
        e.differentiate('a')
    with pytest.raises(RuntimeError):
        e.derivatives
        
    
        
def test_synonyms():
    """ test using synonyms in expression """
    e = ScalarExpression('2 * all', [['a', 'all']])
    assert e.depends_on('a')
    assert not e.depends_on('all')
    


def test_tensor_expression():
    """ test TensorExpression """
    e = TensorExpression('[[0, 1], [2, 3]]')
    assert isinstance(str(e), str)
    assert e.shape == (2, 2)
    assert e.rank == 2
    assert e.constant
    assert e.differentiate('a') == TensorExpression('[[0, 0], [0, 0]]')
    np.testing.assert_allclose(e.value, np.arange(4).reshape(2, 2))
    
    e = TensorExpression('[a, 2*a]')
    assert isinstance(str(e), str)
    assert e.shape == (2,)
    assert e.rank == 1
    assert e.depends_on('a')
    assert not e.constant
    np.testing.assert_allclose(e.differentiate('a').value, np.array([1, 2]))
    with pytest.raises(TypeError):
        e.value
    
    e2 = TensorExpression(e)
    assert isinstance(str(e2), str)
    assert e == e2
    assert e is not e2




import pytest
import numpy as np


@pytest.fixture(scope="session", autouse=True)
def adjust_messages():
    """ helper function adjusting message reporting for all tests """
    # raise all underflow errors
    np.seterr(all='raise', under='ignore')



'''
Defines a vectorial field over a grid

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

from typing import (Callable, Optional, Union, Any, Dict, List, Sequence,
                    TYPE_CHECKING)

import numpy as np
import numba as nb
        

from .base import DataFieldBase
from .scalar import ScalarField
from ..grids.base import GridBase, DimensionError
from ..tools.numba import jit
from ..tools.docstrings import fill_in_docstring

  
if TYPE_CHECKING:
    from ..grids.boundaries.axes import BoundariesData  # @UnusedImport
    from .tensorial import Tensor2Field  # @UnusedImport



class VectorField(DataFieldBase):
    """ Single vector field on a grid
    
    Attributes:
        grid (:class:`~pde.grids.base.GridBase`):
            The underlying grid defining the discretization
        data (:class:`numpy.ndarray`):
            Vector components at the support points of the grid
        label (str):
            Name of the field
    """
    
    rank = 1
    
    
    @classmethod
    def from_scalars(cls, fields: List[ScalarField], label: str = None) \
            -> "VectorField":
        """ create a vector field from a list of ScalarFields
        
        Note that the data of the scalar fields is copied in the process
        
        Args:
            fields (list):
                The list of (compatible) scalar fields
            label (str, optional):
                Name of the returned field
            
        Returns:
            :class:`VectorField`: the resulting vector field
        """
        grid = fields[0].grid
        
        if grid.dim != len(fields):
            raise DimensionError('Grid dimension and number of scalar fields '
                                 f'differ ({grid.dim} != {len(fields)})')
            
        data = []
        for field in fields:
            assert field.grid.compatible_with(grid)
            data.append(field.data)
            
        return cls(grid, data, label)
    

    @classmethod
    @fill_in_docstring
    def from_expression(cls, grid: GridBase, expressions: Sequence[str],
                        label: str = None) -> "VectorField":
        """ create a vector field on a grid from given expressions

        Warning:
            {WARNING_EXEC}
        
        Args:
            grid (:class:`~pde.grids.GridBase`):
                Grid defining the space on which this field is defined
            expressions (list of str):
                A list of mathematical expression, one for each component of the
                vector field. The expressions determine the values as a function
                of the position on the grid. The expressions may contain
                standard mathematical functions and they may depend on the axes
                labels of the grid.
            label (str, optional):
                Name of the field
        """
        from ..tools.expressions import ScalarExpression
        
        if isinstance(expressions, str) or len(expressions) != grid.dim:
            axes_names = grid.axes + grid.axes_symmetric
            raise ValueError(f'Expected {grid.dim} expressions for the '
                             f'coordinates {axes_names}.') 
        
        # obtain the coordinates of the grid points
        points = {name: grid.cell_coords[..., i]
                  for i, name in enumerate(grid.axes)}
        
        # evaluate all vector components at all points 
        data = []
        for expression in expressions:
            expr = ScalarExpression(expression=expression, signature=grid.axes)
            values = np.broadcast_to(expr(**points), grid.shape) 
            data.append(values)

        # create vector field from the data
        return cls(grid=grid,  # lgtm [py/call-to-non-callable]
                   data=data,
                   label=label)
        

    def __getitem__(self, key: int) -> ScalarField:
        """ extract a component of the VectorField """
        if not isinstance(key, int):
            raise IndexError('Index must be an integer')
        return ScalarField(self.grid, self.data[key]) 
                           
     
    def dot(self, other: Union["VectorField", "Tensor2Field"],
            out: Optional[Union[ScalarField, "VectorField"]] = None,
            label: str = 'dot product') -> Union[ScalarField, "VectorField"]:
        """ calculate the dot product involving a vector field
        
        This supports the dot product between two vectors fields as well as the
        product between a vector and a tensor. The resulting fields will be a
        scalar or vector, respectively.
        
        Args:
            other (VectorField or Tensor2Field):
                the second field
            out (ScalarField or VectorField, optional):
                Optional field to which the  result is written.
            label (str, optional):
                Name of the returned field
            
        Returns:
            ScalarField or VectorField: the result of applying the dot operator             
        """
        from .tensorial import Tensor2Field  # @Reimport
        
        # check input
        self.grid.assert_grid_compatible(other.grid)
        if isinstance(other, VectorField):
            result_type = ScalarField
        elif isinstance(other, Tensor2Field):
            result_type = VectorField  # type: ignore
        else:
            raise TypeError('Second term must be a vector or tensor field')

        if out is None:
            out = result_type(self.grid)
        else:
            assert isinstance(out, result_type)
            self.grid.assert_grid_compatible(out.grid)
            
        # calculate the result
        np.einsum('i...,i...->...', self.data, other.data, out=out.data)
        if label is not None:
            out.label = label
            
        return out
    
    __matmul__ = dot  # support python @-syntax for matrix multiplication
    
    
    def get_dot_operator(self) -> Callable:
        """ return operator calculating the dot product involving vector fields
        
        This supports both products between two vectors as well as products
        between a vector and a tensor.
        
        Warning:
            This function does not check types or dimensions.
        
        Returns:
            function that takes two instance of :class:`numpy.ndarray`, which
            contain the discretized data of the two operands. An optional third
            argument can specify the output array to which the result is
            written. Note that the returned function is jitted with numba for
            speed.
        """
        dim = self.grid.dim
        
        @jit
        def inner(a, b, out):
            """ calculate dot product between fields `a` and `b` """
            out[:] = a[0] * b[0]  # overwrite potential data in out
            for i in range(1, dim):
                out[:] += a[i] * b[i]
            return out
        
        if nb.config.DISABLE_JIT:
            def dot(a: np.ndarray, b: np.ndarray, out: np.ndarray = None) \
                    -> np.ndarray:
                """ wrapper deciding whether the underlying function is called
                with or without `out`. """
                if out is None:
                    out = np.empty(b.shape[1:])
                return inner(a, b, out)
            
        else:
            @nb.generated_jit
            def dot(a: np.ndarray, b: np.ndarray, out: np.ndarray = None) \
                    -> np.ndarray:
                """ wrapper deciding whether the underlying function is called
                with or without `out`. """
                if isinstance(a, nb.types.Number):
                    # simple scalar call -> do not need to allocate anything
                    raise RuntimeError('Dot needs to be called with fields')
                
                elif isinstance(out, (nb.types.NoneType, nb.types.Omitted)):
                    # function is called without `out`
                    def f_with_allocated_out(a, b, out):
                        """ helper function allocating output array """
                        return inner(a, b, out=np.empty(b.shape[1:]))
                
                    return f_with_allocated_out
                
                else:
                    # function is called with `out` argument
                    return inner          
        
        return dot
    
    
    def outer_product(self, other: "VectorField", out: "Tensor2Field" = None,
                      label: str = None) -> "Tensor2Field":
        """ calculate the outer product of this vector field with another
        
        Args:
            other (:class:`VectorField`):
                The second vector field
            out (:class:`pde.fields.tensorial.Tensor2Field`, optional):
                Optional tensorial field to which the  result is written.
            label (str, optional):
                Name of the returned field
        
        """
        from .tensorial import Tensor2Field  # @Reimport
        self.assert_field_compatible(other)
        
        if out is None:
            out = Tensor2Field(self.grid)
        else:
            self.grid.assert_grid_compatible(out.grid)
            
        # calculate the result
        np.einsum('i...,j...->ij...', self.data, other.data, out=out.data)
        if label is not None:
            out.label = label
                
        return out
        

    @fill_in_docstring
    def divergence(self, bc: "BoundariesData",
                   out: Optional[ScalarField] = None,
                   label: str = 'divergence') -> ScalarField:
        """ apply divergence operator and return result as a field 
        
        Args:
            bc: 
                The boundary conditions applied to the field.
                {ARG_BOUNDARIES}
            out (ScalarField, optional):
                Optional scalar field to which the  result is written.
            label (str, optional):
                Name of the returned field
            
        Returns:
            ScalarField: the result of applying the operator 
        """
        divergence = self.grid.get_operator('divergence', bc=bc)
        if out is None:
            out = ScalarField(self.grid, divergence(self.data), label=label)
        else:
            assert isinstance(out, ScalarField)
            self.grid.assert_grid_compatible(out.grid)
            divergence(self.data, out=out.data)
        return out
    
    
    @fill_in_docstring
    def gradient(self, bc: "BoundariesData",
                 out: Optional['Tensor2Field'] = None,
                 label: str = 'gradient') -> 'Tensor2Field':
        """ apply (vecotr) gradient operator and return result as a field 
        
        Args:
            bc:
                The boundary conditions applied to the field.
                {ARG_BOUNDARIES}
            out (Tensor2Field, optional):
                Optional tensorial field to which the  result is written.
            label (str, optional):
                Name of the returned field
            
        Returns:
            Tensor2Field: the result of applying the operator 
        """
        vector_gradient = self.grid.get_operator('vector_gradient', bc=bc)
        if out is None:
            from .tensorial import Tensor2Field  # @Reimport
            out = Tensor2Field(self.grid, vector_gradient(self.data),
                               label=label)
        else:
            assert isinstance(out, VectorField)
            self.grid.assert_grid_compatible(out.grid)
            vector_gradient(self.data, out=out.data)
        return out

        
    @fill_in_docstring
    def laplace(self, bc: "BoundariesData",
                out: Optional['VectorField'] = None,
                label: str = 'vector laplacian') -> 'VectorField':
        """ apply vector Laplace operator and return result as a field 
        
        Args:
            bc: 
                The boundary conditions applied to the field.
                {ARG_BOUNDARIES}
            out (VectorField, optional):
                Optional vector field to which the  result is written.
            label (str, optional):
                Name of the returned field
            
        Returns:
            VectorField: the result of applying the operator 
        """
        if out is not None:
            assert isinstance(out, VectorField)
        laplace = self.grid.get_operator('vector_laplace', bc=bc)
        return self.apply(laplace, out=out, label=label)
    
        
    @property
    def integral(self) -> np.ndarray:
        """ :class:`numpy.ndarray`: integral of each component over space """
        return self.grid.integrate(self.data)
                
    
    def to_scalar(self, scalar: str = 'auto',
                  label: Optional[str] = 'scalar `{scalar}`') -> ScalarField:
        """ return a scalar field by applying `method`
        
        The two tensor invariants are given by
        
        Args:
            scalar (str):
                Choose the method to use. Possible  choices are `norm` (the
                default), `max`, `min`, or `squared_sum`.
            label (str, optional):
                Name of the returned field
            
        Returns:
            :class:`pde.fields.scalar.ScalarField`: the scalar field after
            applying the operation
        """
        if scalar == 'auto':
            scalar = 'norm'
            
        if scalar == 'norm':
            data = np.linalg.norm(self.data, axis=0)
            
        elif scalar == 'max':
            data = np.max(self.data, axis=0)
            
        elif scalar == 'min':
            data = np.min(self.data, axis=0)
            
        elif scalar == 'squared_sum' or scalar == 'norm_squared':
            data = np.sum(self.data**2, axis=0)
            
        else:
            raise ValueError(f'Unknown method `{scalar}` for `to_scalar`')
            
        if label is not None:
            label = label.format(scalar=scalar)
            
        return ScalarField(self.grid, data, label=label)

        
    def get_vector_data(self,
                        transpose: bool = False,
                        max_points: int = None,
                        **kwargs) -> Dict[str, Any]:
        r""" return data for a vector plot of the field
        
        Args:
            transpose (bool):
                Determines whether the transpose of the data should be plotted.
            max_points (int):
                The maximal number of points that is used along each axis. This
                option can be used to subsample the data.
            \**kwargs: Additional parameters are forwarded to
                `grid.get_image_data`
        
        Returns:
            dict: Information useful for plotting an vector field        
        """
        # TODO: Handle Spherical and Cartesian grids, too. This could be
        # implemented by adding a get_vector_data method to the grids
        if self.grid.dim == 2:
            vx = self[0].get_image_data(**kwargs)
            vy = self[1].get_image_data(**kwargs)
            data = vx  # use one of the fields to extract basic information
            data['data_x'] = vx.pop('data')
            data['data_y'] = vy['data']
            data['title'] = self.label
            
        else:
            raise NotImplementedError()
        
        # transpose the data if requested
        if transpose:
            data['x'], data['y'] = data['y'], data['x']
            data['data_x'], data['data_y'] = data['data_y'].T, data['data_x'].T
            data['label_x'], data['label_y'] = data['label_y'], data['label_x']        

        # reduce the sampling of the vector points
        if max_points is not None:
            shape = data['data_x'].shape
            for axis, size in enumerate(shape):
                if size > max_points:
                    # sub-sample the data
                    idx_f = np.linspace(0, size - 1, max_points)
                    idx_i = np.round(idx_f).astype(int)
                    #
                    data['data_x'] = np.take(data['data_x'], idx_i, axis=axis)
                    data['data_y'] = np.take(data['data_y'], idx_i, axis=axis)
                    if axis == 0:
                        data['y'] = data['y'][idx_i]
                    elif axis == 1:
                        data['x'] = data['x'][idx_i]
        
        return data    



'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import pytest
import numpy as np

from .. import MemoryStorage
from ...grids import UnitGrid
from ...fields import ScalarField, VectorField, Tensor2Field, FieldCollection


            
def test_memory_storage():
    """ test methods specific to memory storage """
    sf = ScalarField(UnitGrid([1]))
    s1 = MemoryStorage()
    s1.start_writing(sf)
    s1.append(np.array([0]), 0)
    s1.append(np.array([2]), 1)
    s2 = MemoryStorage()
    s2.start_writing(sf)
    s2.append(np.array([1]), 0)
    s2.append(np.array([3]), 1)
    
    s3 = MemoryStorage.from_collection([s1, s2])
    np.testing.assert_allclose(np.ravel(s3.data), np.arange(4))



def test_field_type_guessing():
    """ test the ability to guess the field type """
    for cls in [ScalarField, VectorField, Tensor2Field]:
        grid = UnitGrid([3])
        field = cls.random_normal(grid)
        s = MemoryStorage()
        s.start_writing(field)
        s.append(field.data, 0)
        s.append(field.data, 1)
        
        # delete information
        s._field = None
        s.info = {}
        
        assert len(s) == 2
        assert s[0] == field
        
    field = FieldCollection([ScalarField(grid), VectorField(grid)])
    s = MemoryStorage()
    s.start_writing(field)
    s.append(field.data, 0)
    
    # delete information
    s._field = None
    s.info = {}
    
    with pytest.raises(RuntimeError):
        s[0]

    


"""
Stochastic simulation
=====================

This example illustrates how a stochastic simulation can be done.
"""

from pde import (KPZInterfacePDE, UnitGrid, ScalarField, MemoryStorage,
                 plot_kymograph)


grid = UnitGrid([32])  # generate grid
state = ScalarField.random_harmonic(grid)  # generate initial condition

eq = KPZInterfacePDE(noise=1)  # define the SDE
storage = MemoryStorage()
eq.solve(state, t_range=10, dt=0.01, tracker=storage.tracker(1))
plot_kymograph(storage, show=True)



"""
The py-pde package provides classes and methods for solving partial differential
equations.
"""

__version__ = "0.8.1"

from .fields import *  # @UnusedWildImport
from .grids import *  # @UnusedWildImport
from .pdes import *  # @UnusedWildImport
from .solvers import *  # @UnusedWildImport
from .storage import *  # @UnusedWildImport
from .trackers import *  # @UnusedWildImport
from .visualization import *  # @UnusedWildImport

from .tools.misc import environment
from .tools.parameters import Parameter



'''
Functions and classes for plotting simulation data

.. autosummary::
   :nosignatures:

   ScalarFieldPlot
   plot_magnitudes
   plot_kymograph
   
   
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import logging
import warnings
import time
from typing import (Union, Callable, Optional, Any, Dict, List, Tuple)

import numpy as np
import matplotlib.pyplot as plt

from ..fields import FieldCollection
from ..fields.base import FieldBase, DataFieldBase
from ..storage.base import StorageBase
from ..tools.misc import display_progress
from ..tools.docstrings import fill_in_docstring
from ..tools.plotting import finalize_plot


ScaleData = Union[str, float, Tuple[float, float]]



def _add_horizontal_colorbar(im, ax, num_loc: int = 5) -> None:
    """ adds a horizontal colorbar for image `im` to the axis `ax`
    
    Args:
        im: The result of calling :func:`matplotlib.pyplot.imshow`
        ax: The matplotlib axes to which the colorbar is added
        num_loc (int): Number of ticks
    """
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    from matplotlib.ticker import MaxNLocator
    
    fig = ax.figure
    divider = make_axes_locatable(ax)
    cax = divider.new_vertical(size="5%", pad=0.1, pack_start=True)
    fig.add_axes(cax)
    cb = fig.colorbar(im, cax=cax, orientation="horizontal")
    cb.locator = MaxNLocator(num_loc)
    cb.update_ticks()



def extract_field(fields: FieldBase,
                  source: Union[None, int, Callable] = None,
                  check_rank: Optional[int] = None) -> DataFieldBase:
    """Extracts a single field from a possible collection.
    
    Args:
        fields (:class:`~pde.fields.FieldBase`):
            The field from which data is extracted
        source (int or callable, optional):
            Determines how a field is extracted from `fields`. If `None`,
            `fields` is passed as is, assuming it is already a scalar field.
            This works for the simple, standard case where only a single
            :class:`~pde.fields.scalar.ScalarField` is treated. Alternatively,
            `source` can be an integer, indicating which field is extracted from
            an instance of :class:`~pde.fields.FieldCollection`.
            Lastly, `source` can be a function that takes `fields` as an
            argument and returns the desired field.
        check_rank (int, optional):
            Can be given to check whether the extracted field has the correct
            rank (0 = ScalarField, 1 = VectorField, ...).
            
    Returns:
        :class:`~pde.fields.DataFieldBase`: The extracted field
    """
    # extract the source
    if source is None:
        field = fields
    elif callable(source):
        field = source(fields)
    else:
        if isinstance(fields, FieldCollection):
            field = fields[source]
        else:
            raise TypeError(f'Cannot extract component {source} from instance '
                            f'of {fields.__class__.__name__}')
        
    if isinstance(field, FieldCollection):
        raise RuntimeError('Extract field is a collection')
    
    # test whether the rank is as expected
    if check_rank is not None and field.rank != check_rank:  # type: ignore
        raise RuntimeError(f'Rank of extracted field is not {check_rank}')
        
    return field  # type: ignore



class ScalarFieldPlot():
    """ class managing compound plots of scalar fields """
    
    
    @fill_in_docstring
    def __init__(self,                 
                 fields: FieldBase,
                 quantities=None,
                 scale: ScaleData = 'automatic',
                 fig=None,
                 title: Optional[str] = None,
                 tight: bool = False,
                 show: bool = True):
        """
        Args:
            fields (:class:`~pde.fields.base.FieldBase`):
                Collection of fields
            quantities:
                {ARG_PLOT_QUANTITIES}
            scale (str, float, tuple of float):
                {ARG_PLOT_SCALE}
            fig (:class:`matplotlib.figure.Figure):
                Figure to be used for plotting. If `None`, a new figure is
                created.
            title (str):
                Title of the plot.
            tight (bool):
                Whether to call :func:`matplotlib.pyplot.tight_layout`. This
                affects the layout of all plot elements.
            show (bool):
                Flag determining whether to show a plot. If `False`, the plot is
                kept in the background, which can be useful if it only needs to
                be written to a file.
        """
        self.grid = fields.grid
        self.quantities = self._prepare_quantities(fields, quantities,
                                                   scale=scale)
        self.show = show
             
        # figure out whether plots are shown in jupyter notebook
        try:
            from ipywidgets import Output
        except ImportError:
            ipython_plot = False
        else:
            ipython_plot = self.show
            
        if ipython_plot:
            # plotting is done in an ipython environment using widgets
            from IPython.display import display
            self._ipython_out = Output()
            with self._ipython_out:
                self._initialize(fields, scale, fig, title, tight)
            display(self._ipython_out)
            
        else:
            # plotting is done using a simple matplotlib backend
            self._ipython_out = None
            self._initialize(fields, scale, fig, title, tight)
            
        if self.show:
            self._show()

    
    @classmethod
    @fill_in_docstring
    def from_storage(cls, storage: StorageBase,
                     quantities=None,
                     scale: ScaleData = 'automatic',
                     tight: bool = False,
                     show: bool = True) -> "ScalarFieldPlot":
        """ create ScalarFieldPlot from storage
        
        Args:
            storage (:class:`~pde.storage.base.StorageBase`):
                Instance of the storage class that contains the data
            quantities:
                {ARG_PLOT_QUANTITIES}
            scale (str, float, tuple of float):
                {ARG_PLOT_SCALE}
            tight (bool):
                Whether to call :func:`matplotlib.pyplot.tight_layout`. This
                affects the layout of all plot elements.
            show (bool):
                Flag determining whether to show a plot. If `False`, the plot is
                kept in the background.
            
        Returns:
            :class:`~pde.visualization.plotting.ScalarFieldPlot`
        """
        fields = storage.get_field(0)
        quantities = cls._prepare_quantities(fields, quantities=quantities,
                                             scale=scale)
        
        # resolve automatic scaling
        for quantity_row in quantities:
            for quantity in quantity_row:
                if quantity.get('scale', 'automatic') == 'automatic':
                    vmin, vmax = +np.inf, -np.inf
                    for data in storage:
                        field = extract_field(data, quantity.get('source'))
                        img = field.get_image_data()
                        vmin = np.nanmin([np.nanmin(img['data']), vmin])
                        vmax = np.nanmax([np.nanmax(img['data']), vmax])
                    quantity['scale'] = (vmin, vmax)
            
        # actually setup 
        return cls(fields,  # lgtm [py/call-to-non-callable]
                   quantities,
                   tight=tight, show=show)
        
        
    @staticmethod
    @fill_in_docstring
    def _prepare_quantities(fields: FieldBase,
                            quantities,
                            scale: ScaleData = 'automatic') \
            -> List[List[Dict[str, Any]]]:
        """ internal method to prepare quantities
        
        Args:
            fields (:class:`~pde.fields.base.FieldBase`):
                The field containing the data to show
            quantities (dict):
                {ARG_PLOT_QUANTITIES}
            scale (str, float, tuple of float):
                {ARG_PLOT_SCALE}
        
        Returns:
            list of list of dict: a 2d arrangements of panels that define what
                quantities are shown.
        """
        if quantities is None:
            # show each component by default
            if isinstance(fields, FieldCollection):
                quantities = []
                for i, field in enumerate(fields):
                    title = field.label if field.label else f'Field {i + 1}'
                    quantity = {'title': title, 'source': i, 'scale': scale}
                    quantities.append(quantity)
            else:
                quantities = [{'title': 'Concentration', 'source': None}]
                
        # make sure panels is a 2d array
        if isinstance(quantities, dict):
            quantities = [[quantities]]
        elif isinstance(quantities[0], dict):
            quantities = [quantities]
            
        # set the scaling for quantities where it is not yet set
        for row in quantities:
            for col in row:
                col.setdefault('scale', scale)
            
        return quantities  # type: ignore
        
        
    def __del__(self):
        try:
            if hasattr(self, 'fig') and self.fig:
                plt.close(self.fig)
        except Exception:
            pass  # can occur when shutting down python interpreter

    
    @fill_in_docstring
    def _initialize(self, fields: FieldBase,
                    scale: ScaleData = 'automatic',
                    fig=None,
                    title: Optional[str] = None,
                    tight: bool = False):         
        """ initialize the plot creating the figure and the axes
        
        Args:
            fields (:class:`~pde.fields.base.FieldBase`):
                Collection of fields
            scale (str, float, tuple of float):
                {ARG_PLOT_SCALE}
            fig (:class:`matplotlib.figure.Figure):
                Figure to be used for plotting. If `None`, a new figure is
                created.
            title (str):
                Title of the plot.
            tight (bool):
                Whether to call :func:`matplotlib.pyplot.tight_layout`. This
                affects the layout of all plot elements.
        """        
        import matplotlib.cm as cm
        
        num_rows = len(self.quantities)
        num_cols = max(len(p) for p in self.quantities)
             
        example_image = fields.get_image_data()
        
        # set up the figure
        if fig is None:
            figsize = (4 * num_cols, 4 * num_rows)
            self.fig, self.axes = plt.subplots(num_rows, num_cols, sharey=True,
                                               squeeze=False, figsize=figsize)
        else:
            self.fig = fig
            self.axes = fig.axes
            
            
        if title is None:
            title = ""        
        self.sup_title = plt.suptitle(title)
        
        # set up all images
        empty = np.zeros_like(example_image['data'])

        self.images: List[List[Any]] = []
        for i, panel_row in enumerate(self.quantities):
            img_row = []
            for j, panel in enumerate(panel_row):
                # determine scale of the panel
                panel_scale = panel.get('scale', scale)
                if panel_scale == 'automatic':
                    vmin, vmax = None, None
                elif panel_scale == 'unity':
                    vmin, vmax = 0, 1
                elif panel_scale == 'symmetric':
                    vmin, vmax = -1, 1
                else:
                    try:
                        vmin, vmax = panel_scale
                    except TypeError:
                        vmin, vmax = 0, panel_scale
                    
                # determine colormap
                cmap = panel.get('cmap')
                if cmap is None:
                    if vmin is None or vmax is None:
                        cmap = cm.viridis
                    elif np.isclose(-vmin, vmax):
                        cmap = cm.coolwarm
                    elif np.isclose(vmin, 0):
                        cmap = cm.gray
                    else:
                        cmap = cm.viridis

                # initialize image    
                ax = self.axes[i, j]
                img = ax.imshow(empty, vmin=vmin, vmax=vmax, cmap=cmap,
                                extent=example_image['extent'], origin='lower',
                                interpolation='quadric')
                _add_horizontal_colorbar(img, ax)
                ax.set_title(panel.get('title', ''))
                ax.set_xlabel(example_image['label_x'])
                ax.set_ylabel(example_image['label_y'])
                
                # store the default value alongside
                img._default_vmin_vmax = vmin, vmax
                img_row.append(img)
            self.images.append(img_row)
            
        if tight:
            # adjust layout and leave some room for title
            self.fig.tight_layout(rect=[0, 0.03, 1, 0.95])

            
    def _update_data(self, fields: FieldBase,
                     title: Optional[str] = None) -> None:
        """ update the fields in the current plot
        
        Args:
            fields (:class:`~pde.fields.base.FieldBase`):
                The field or field collection of which the defined quantities
                are shown.
            title (str, optional):
                The title of this view. If `None`, the current title is not
                changed. 
        """
        assert isinstance(fields, FieldBase)
        
        if title:
            self.sup_title.set_text(title)
        
        # iterate over all panels and update their content
        for i, panel_row in enumerate(self.quantities):
            for j, panel in enumerate(panel_row):
                # obtain image data
                field = extract_field(fields, panel.get('source'))
                img_data = field.get_image_data()
                
                # set the data in the correct panel
                img = self.images[i][j]
                img.set_data(img_data['data'])
                vmin, vmax = img._default_vmin_vmax
                if vmin is None:
                    vmin = img_data['data'].min()
                if vmax is None:
                    vmax = img_data['data'].max()
                img.set_clim(vmin, vmax)
                    
            
    def _show(self):
        """ show the updated plot """
        if self._ipython_out:
            # seems to be in an ipython instance => update widget
            from IPython.display import display, clear_output
            with self._ipython_out:
                clear_output(wait=True)
                display(self.fig)
                
            # add a small pause to allow the GUI to run it's event loop
            time.sleep(0.01)
            
        else:
            # seems to be in a normal matplotlib window => update it
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                # add a small pause to allow the GUI to run it's event loop
                plt.pause(0.01)
        
        
    def update(self, fields: FieldBase, title: Optional[str] = None) -> None:
        """ update the plot with the given fields
        
        Args:
            fields:
                The field or field collection of which the defined quantities
                are shown.
            title (str, optional):
                The title of this view. If `None`, the current title is not
                changed. 
        """
        self._update_data(fields, title)
        if self.show:
            self._show()

            
    def savefig(self, path: str, **kwargs):
        """ save plot to file 
        
        Args:
            path (str):
                The path to the file where the image is written. The file
                extension determines the image format
            **kwargs:
                Additional arguments are forwarded to 
                :meth:`matplotlib.figure.Figure.savefig`.
        """
        self.fig.savefig(path, **kwargs)
        
    
    def make_movie(self, storage: StorageBase,
                   filename: str,
                   progress: bool = True) -> None:
        """ make a movie from the data stored in storage
        
        Args:
            storage (:class:`~pde.storage.base.StorageBase`):
                The storage instance that contains all the data for the movie
            filename (str):
                The filename to which the movie is written. The extension
                determines the format used.
            progress (bool):
                Flag determining whether the progress of making the movie is
                shown.
        """
        from ..visualization.movies import Movie

        # create the iterator over the data
        if progress:
            data_iter = display_progress(storage.items(), total=len(storage))
        else:
            data_iter = storage.items()
            
        with Movie(filename=filename, verbose=False) as movie:
            # iterate over all time steps
            for t, data in data_iter:
                self.update(data, title=f'Time {t:g}')
                movie.add_figure(self.fig)



@fill_in_docstring
def plot_magnitudes(storage: StorageBase,
                    filename: str = None,
                    quantities=None,
                    ax=None,
                    title: str = None,
                    show: bool = False,
                    **kwargs) -> None:
    r""" create a plot of spatially integrated quantities from a given storage.
    
    Note that the plot will not be displayed when a filename is given and no
    plot axes are specified.
    
    Args:
        storage:
            Instance of :class:`~pde.storage.base.StorageBase` that contains
            the simulation data that will be plotted
        filename (str):
            If given, the resulting image is written to this file.
        quantities:
            {ARG_PLOT_QUANTITIES}
        ax:
            Matplotlib figure axes to be used for plotting. If `None`, a new
            figure is created
        title (str):
            Determines the title of the figure
        show (bool):
            Flag determining whether :func:`matplotlib.pyplot.show` is called
        \**kwargs:
            All remaining parameters are forwarded to the `ax.plot` method
    """
    if quantities is None:
        fields = storage.get_field(0)
        if fields.label:
            label_base = fields.label 
        else:
            label_base = kwargs.get('label', 'Field')
        
        if isinstance(fields, FieldCollection):
            quantities = []
            for i, field in enumerate(fields):
                if field.label:
                    label = field.label 
                else:
                    label = label_base + f' {i + 1}'
                quantities.append({'label': label, 'source': i})
        else:
            quantities = [{'label': label_base, 'source': None}]
            
    logging.getLogger(__name__).debug('Quantities: %s', quantities)
    
    # prepare data field
    data = [{'label': quantity['label'], 'values': []}
            for quantity in quantities]
    
    # calculate the data
    for fields in storage:
        for i, quantity in enumerate(quantities):
            source = quantity['source']
            if callable(source):
                value = source(fields)
            elif source is None:
                value = fields.magnitude  # type: ignore
            else:
                value = fields[source].magnitude  # type: ignore
            value *= quantity.get('scale', 1)
            data[i]['values'].append(value)
    
    close_figure = (bool(filename) and ax is None)
    if ax is None:
        # create new figure
        ax = plt.figure().gca()
    
    # plot the data
    for d in data:
        kwargs['label'] = d['label']
        ax.plot(storage.times, d['values'], **kwargs)
        
    ax.set_xlabel('Time')
    if len(data) == 1:
        ax.set_ylabel(kwargs['label'])
        
    if len(data) > 1:
        ax.legend(loc='best')
    
    finalize_plot(ax, title=title, filename=filename, show=show,
                  close_figure=close_figure)
            


def plot_kymograph(storage: StorageBase,
                   filename: str = None,
                   scalar: str = 'auto',
                   extract: str = 'auto',
                   colorbar: bool = True,
                   transpose: bool = False,
                   title: str = None,
                   show: bool = False,
                   ax=None,
                   **kwargs):
    r""" plots a simple kymograph from stored data
    
    The kymograph shows line data stacked along time. Consequently, the
    resulting image shows space along the horizontal axis and time along the
    vertical axis.
    
    Note that the plot will not be displayed when a filename is given and no
    plot axes are specified.
    
    Args:
        storage (:class:`~droplets.simulation.storage.StorageBase`):
            The storage instance that contains all the data for the movie
        filename (str):
            If given, the resulting image is written to this file.
        scalar (str):
            The method for extracting scalars as described in
            :meth:`DataFieldBase.to_scalar`.
        extract (str):
            The method used for extracting the line data. See the docstring
            of the grid method `get_line_data` to find supported values.
        colorbar (bool):
            Whether to show a colorbar or not
        transpose (bool):
            Determines whether the transpose of the data should is plotted
        title (str):
            Determines the title of the figure
        show (bool):
            Flag determining whether :func:`matplotlib.pyplot.show` is called
        ax: Figure axes to be used for plotting. If `None`, a new figure is
            created
        \**kwargs:
            Additional keyword arguments are passed to
            :func:`matplotlib.pyplot.imshow`.
            
    Returns:
        Result of :func:`matplotlib.pyplot.imshow`
    """
    close_figure = (bool(filename) and ax is None)
    if ax is None:
        ax = plt.figure().gca()  # create new figure

    full_data = []
    for _, data in storage.items():
        img_data = data.get_line_data(scalar=scalar, extract=extract)
        full_data.append(img_data['data_y'])
        
    full_data = np.array(full_data)
    extent = np.r_[img_data['extent_x'], storage.times[0], storage.times[-1]]
    
    if transpose:
        full_data = full_data.T  # type: ignore
        label_x, label_y = 'Time', img_data['label_x']
    else:
        label_x, label_y = img_data['label_x'], 'Time'
    
    res = ax.imshow(full_data, extent=extent, origin='lower', **kwargs)
    
    ax.set_xlabel(label_x)
    ax.set_ylabel(label_y)
    ax.set_aspect('auto') 

    if colorbar:
        from ..tools.misc import add_scaled_colorbar
        add_scaled_colorbar(res, ax=ax)
        
    if title is None:
        title = img_data['label_y']

    finalize_plot(ax, title=title, filename=filename, show=show,
                  close_figure=close_figure)
    
    return res

    


'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''


import numpy as np

from ..math import SmoothData1D



def test_SmoothData1D():
    """ test smoothing """
    x = np.random.uniform(0, 1, 128)
    xs = np.linspace(0, 1, 16)[1:-1]

    s = SmoothData1D(x, np.ones_like(x))
    np.testing.assert_allclose(s(xs), 1)

    s = SmoothData1D(x, x)
    np.testing.assert_allclose(s(xs), xs, atol=0.1)

    s = SmoothData1D(x, np.sin(x))
    np.testing.assert_allclose(s(xs), np.sin(xs), atol=0.1)



'''
Auxiliary mathematical functions

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

from typing import Tuple

import numpy as np


    
class SmoothData1D():
    """ provide structure for smoothing data in 1d
    
    The data is given a pairs of `x` and `y`, the assumption being that there is
    an underlying relation `y_n = f(x_n)`.
    """
    
    sigma_auto_scale: float = 10  # scale for setting automatic values for sigma
    
    def __init__(self, x, y, sigma: float = None):
        """ initialize with data
        
        Args:
            x: List of x values
            y: List of y values
            sigma (float): The size of the smoothing window using units of `x`.
                If it is not given, the average distance of x values multiplied
                by `sigma_auto_scale` is used. 
        """
        self.x = np.ravel(x)
        self.y = np.ravel(y)
        if self.x.shape != self.y.shape:
            raise ValueError('`x` and `y` must have equal number of elements')
            
        if sigma is None:
            self.sigma = self.sigma_auto_scale * self.x.ptp() / len(self.x)
        else:
            self.sigma = sigma
        
        
    @property
    def bounds(self) -> Tuple[float, float]:
        """ return minimal and maximal `x` values """
        return self.x.min(), self.x.max()
        
        
    def __call__(self, xs):
        """ return smoothed y values for the positions given in `xs`
        
        Args:
            xs: a list of x-values
            
        Returns:
            :class:`numpy.ndarray`: The associated y-values
        """
        xs = np.asanyarray(xs)
        shape = xs.shape
        xs = np.ravel(xs)
        scale = 0.5 * self.sigma**-2
        
        with np.errstate(under='ignore'):
            weight = np.exp(-scale * (self.x[:, None] - xs[None, :])**2)
            weight /= weight.sum(axis=0)
        result = np.dot(self.y, weight)
        return result.reshape(shape)    
    


'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

from ..parse_duration import parse_duration



def test_parse_duration():
    """ test function signature checks """
    def p(value):
        return parse_duration(value).total_seconds()
    assert p('0') == 0
    assert p('1') == 1
    assert p('1:2') == 62
    assert p('1:2:3') == 3723



'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import numpy as np

import pytest

from ... import UnitGrid, CartesianGrid, ScalarField
from .. import solve_laplace_equation, solve_poisson_equation



def test_pde_poisson_solver_1d():
    """ test the poisson solver on 1d grids """
    # solve Laplace's equation
    grid = UnitGrid([4])
    res = solve_laplace_equation(grid, bc=[{'value': -1}, {'value': 3}])
    np.testing.assert_allclose(res.data, grid.axes_coords[0] - 1)
    
    res = solve_laplace_equation(grid, bc=[{'value': -1}, {'derivative': 1}])
    np.testing.assert_allclose(res.data, grid.axes_coords[0] - 1)

    # test Poisson equation with 2nd Order BC
    res = solve_laplace_equation(grid, bc=[{'value': -1}, 'extrapolate'])
    
    # solve Poisson's equation
    grid = CartesianGrid([[0, 1]], 4)
    field = ScalarField(grid, data=1)
    
    res = solve_poisson_equation(field, bc=[{'value': 1}, {'derivative': 1}])
    xs = grid.axes_coords[0]
    np.testing.assert_allclose(res.data, 1 + 0.5 * xs**2, rtol=1e-2)
    
    # test inconsistent problem
    field.data = 1
    with pytest.raises(RuntimeError, match='Neumann'):
        solve_poisson_equation(field, {'derivative': 0})
    


def test_pde_poisson_solver_2d():
    """ test the poisson solver on 2d grids """
    grid = CartesianGrid([[0, 2 * np.pi]] * 2, 16)
    bcs = [{'value': 'sin(y)'}, {'value': 'sin(x)'}]
    
    # solve Laplace's equation
    res = solve_laplace_equation(grid, bcs)
    xs = grid.cell_coords[..., 0]
    ys = grid.cell_coords[..., 1]
    
    # analytical solution was obtained with Mathematica
    expect = (np.cosh(np.pi - ys) * np.sin(xs) +
              np.cosh(np.pi - xs) * np.sin(ys)) / np.cosh(np.pi)
    np.testing.assert_allclose(res.data, expect, atol=1e-2, rtol=1e-2)
    
    # test more complex case for exceptions
    bcs = [{'value': 'sin(y)'}, {'curvature': 'sin(x)'}]
    res = solve_laplace_equation(grid, bc=bcs)
    
    
    


"""
Defines an implicit solver
   
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de> 
"""

from typing import Callable, Tuple

import numpy as np
import numba as nb

from .base import SolverBase
from ..pdes.base import PDEBase
from ..fields.base import FieldBase
from ..tools.numba import jit



class ConvergenceError(RuntimeError):
    pass



class ImplicitSolver(SolverBase):
    """ class for solving partial differential equations implicitly """

    name = 'implicit'


    def __init__(self, pde: PDEBase,
                 maxiter: int = 100,
                 maxerror: float = 1e-4,
                 backend: str = 'auto'):
        """ initialize the implicit solver
        
        Args:
            pde (:class:`~pde.pdes.base.PDEBase`):
                The instance describing the pde that needs to be solved
            maxiter (int):
                The maximal number of iterations per step
            maxerror (float):
                The maximal error that is permitted in each step
            backend (str):
                Determines how the function is created. Accepted  values are
                'numpy` and 'numba'. Alternatively, 'auto' lets the code decide
                for the most optimal backend.
        """        
        super().__init__(pde)
        self.maxiter = maxiter
        self.maxerror = maxerror
        self.backend = backend
    

    def make_stepper(self, state: FieldBase, dt=None) -> Callable:
        """ return a stepper function using an implicit scheme
        
        Args:
            state (:class:`~pde.fields.FieldBase`):
                An example for the state from which the grid and other
                information can be extracted
            dt (float):
                Time step of the explicit stepping. If `None`, this solver
                specifies 1e-3 as a default value.
                
        Returns:
            Function that can be called to advance the `state` from time
            `t_start` to time `t_end`. The function call signature is
            `(state: numpy.ndarray, t_start: float, t_end: float)`        
        """
        # support `None` as a default value, so the controller can signal that
        # the solver should use a default time step
        if dt is None:
            dt = 1e-3
        
        self.info['dt'] = dt
        self.info['steps'] = 0
        self.info['function_evaluations'] = 0
        self.info['scheme'] = 'implicit-euler'
        self.info['stochastic'] = False
    
        rhs = self._make_pde_rhs(state, backend=self.backend,
                                 allow_stochastic=False)
        maxiter = int(self.maxiter)
        maxerror2 = self.maxerror ** 2
        
        # handle deterministic  version of the pde
        def inner_stepper(state_data: np.ndarray, t_start: float, steps: int) \
                -> Tuple[float, int]:
            """ compiled inner loop for speed """
            nfev = 0  # count function evaluations
            for i in range(steps):
                t = t_start + i * dt  # current time point
                tn = t + dt           # next time point
                
                # estimate state at next time point 
                evolution_last = dt * rhs(state_data, t)

                for n in range(maxiter):
                    # fixed point iteration for improving state after dt
                    state_guess = state_data + evolution_last
                    evolution_this = dt * rhs(state_guess, tn)
                    
                    # calculate mean squared error
                    err = 0
                    for j in range(state_data.size):
                        err += (state_guess.flat[j] - state_data.flat[j] -
                                evolution_this.flat[j])**2
                    err /= state_data.size
                                
                    if err < maxerror2:
                        # fix point iteration converged
                        break
                    
                    evolution_last = evolution_this
                else:
                    with nb.objmode:
                        self._logger.warn('Implicit Euler step did not '
                                          'converge after %d iterations at '
                                          't=%g (error=%g)', maxiter, t, err)
                    raise ConvergenceError('Implicit Euler step did not '
                                           'converge.')
                nfev += n + 1
                state_data += evolution_this
                     
            return tn, nfev
        
        if self.info['backend'] == 'numba':
            # compile inner step
            inner_stepper = jit(inner_stepper)
        
        def stepper(state: FieldBase, t_start: float, t_end: float) \
                -> float:
            """ use Euler stepping to advance `state` from `t_start` to
            `t_end` """
            # calculate number of steps (which is at least 1)
            steps = max(1, int(np.ceil((t_end - t_start) / dt)))
            t_last, nfev = inner_stepper(state.data, t_start, steps)
            self.info['steps'] += steps
            self.info['function_evaluations'] += nfev
            return t_last        
            
        self._logger.info(f'Initialized implicit Euler stepper with dt=%g', dt)
        return stepper
        


"""
Defines an explicit solver supporting various methods
   
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de> 
"""

from typing import Callable

import numpy as np

from .base import SolverBase
from ..pdes.base import PDEBase 
from ..fields.base import FieldBase
from ..tools.numba import jit



class ExplicitSolver(SolverBase):
    """ class for solving partial differential equations explicitly """

    name = 'explicit'


    def __init__(self, pde: PDEBase,
                 scheme: str = 'euler',
                 backend: str = 'auto'):
        """ initialize the explicit solver
        
        Args:
            pde (:class:`~pde.pdes.base.PDEBase`):
                The instance describing the pde that needs to be solved
            scheme (str):
                Defines the explicit scheme to use. Supported values are
                'euler', 'runge-kutta' (or 'rk' for short).
            backend (str):
                Determines how the function is created. Accepted  values are
                'numpy` and 'numba'. Alternatively, 'auto' lets the code decide
                for the most optimal backend.
        """        
        super().__init__(pde)
        self.scheme = scheme
        self.backend = backend
    
    
    def _make_euler_stepper(self, state: FieldBase, dt: float) -> Callable:
        """ make a simple Euler stepper
        
        Args:
            state (:class:`~pde.fields.FieldBase`):
                An example for the state from which the grid and other
                information can be extracted
            dt (float):
                Time step of the explicit stepping. If `None`, this solver
                specifies 1e-3 as a default value.
                
        Returns:
            Function that can be called to advance the `state` from time
            `t_start` to time `t_end`. The function call signature is
            `(state: numpy.ndarray, t_start: float, t_end: float)`            
        """
        rhs = self._make_pde_rhs(state, backend=self.backend,
                                 allow_stochastic=True)
        
        if self.pde.is_sde:
            # handle stochastic version of the pde
            
            def stepper(state_data: np.ndarray, t_start: float, steps: int) \
                    -> float:
                """ compiled inner loop for speed """
                for i in range(steps):
                    # calculate the right hand side
                    t = t_start + i * dt
                    evolution_rate, noise_realization = rhs(state_data, t)
                    state_data += dt * evolution_rate
                    if noise_realization is not None:
                        state_data += np.sqrt(dt) * noise_realization
                return t + dt        

            self.info['stochastic'] = True
            self._logger.info(f'Initialized explicit Euler-Maruyama stepper '
                              'with dt=%g', dt)
            
        else:
            # handle deterministic  version of the pde
            def stepper(state_data: np.ndarray, t_start: float, steps: int) \
                    -> float:
                """ compiled inner loop for speed """
                for i in range(steps):
                    # calculate the right hand side
                    t = t_start + i * dt
                    state_data += dt * rhs(state_data, t)
                return t + dt
        
            self.info['stochastic'] = False
            self._logger.info(f'Initialized explicit Euler stepper with dt=%g',
                              dt)
            
        return stepper 
    
    
    def _make_rk45_stepper(self, state: FieldBase, dt: float) -> Callable:
        """ make a simple Euler stepper
        
        Args:
            state (:class:`~pde.fields.FieldBase`):
                An example for the state from which the grid and other
                information can be extracted
            dt (float):
                Time step of the explicit stepping. If `None`, this solver
                specifies 1e-3 as a default value.
                
        Returns:
            Function that can be called to advance the `state` from time
            `t_start` to time `t_end`. The function call signature is
            `(state: numpy.ndarray, t_start: float, t_end: float)`            
        """
        rhs = self._make_pde_rhs(state, backend=self.backend,
                                 allow_stochastic=False)
        self.info['stochastic'] = False
            
        
        def stepper(state_data: np.ndarray, t_start: float, steps: int) \
                -> float:
            """ compiled inner loop for speed """
            for i in range(steps):
                # calculate the right hand side
                t = t_start + i * dt
                
                # calculate the intermediate values in Runge-Kutta
                k1 = dt * rhs(state_data, t)
                k2 = dt * rhs(state_data + 0.5 * k1, t + 0.5 * dt)
                k3 = dt * rhs(state_data + 0.5 * k2, t + 0.5 * dt)
                k4 = dt * rhs(state_data + k3, t + dt)  
                
                state_data += (k1 + 2 * k2 + 2 * k3 + k4) / 6
            
            return t + dt        
        
        self._logger.info(f'Initialized explicit Runge-Kutta-45 stepper with '
                          'dt=%g', dt)
        return stepper 


    def make_stepper(self, state: FieldBase, dt=None) -> Callable:
        """ return a stepper function using an explicit scheme
        
        Args:
            state (:class:`~pde.fields.FieldBase`):
                An example for the state from which the grid and other
                information can be extracted
            dt (float):
                Time step of the explicit stepping. If `None`, this solver
                specifies 1e-3 as a default value.
                
        Returns:
            Function that can be called to advance the `state` from time
            `t_start` to time `t_end`. The function call signature is
            `(state: numpy.ndarray, t_start: float, t_end: float)`        
        """
        # support `None` as a default value, so the controller can signal that
        # the solver should use a default time step
        if dt is None:
            dt = 1e-3
        
        self.info['dt'] = dt
        self.info['steps'] = 0
        self.info['scheme'] = self.scheme
        
        if self.scheme == 'euler':
            inner_stepper = self._make_euler_stepper(state, dt)
        elif self.scheme in {'runge-kutta', 'rk', 'rk45'}:
            inner_stepper = self._make_rk45_stepper(state, dt)
        else:
            raise ValueError(f"Explicit scheme {self.scheme} is not supported")

        if self.info['backend'] == 'numba':
            # compile inner step
            inner_stepper = jit(inner_stepper)
        
        def stepper(state: FieldBase, t_start: float, t_end: float) \
                -> float:
            """ use Euler stepping to advance `state` from `t_start` to
            `t_end` """
            # calculate number of steps (which is at least 1)
            steps = max(1, int(np.ceil((t_end - t_start) / dt)))
            t_last = inner_stepper(state.data, t_start, steps)
            self.info['steps'] += steps
            return t_last  # type: ignore        
            
        return stepper

        


'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import pytest
import numpy as np

from .. import WavePDE
from ...grids import UnitGrid
from ...fields import ScalarField


@pytest.mark.parametrize('dim', [1, 2])
def test_wave_consistency(dim):
    """ test some methods of the wave model """
    eq = WavePDE()
    assert isinstance(str(eq), str)
    assert isinstance(repr(eq), str)
    
    grid = UnitGrid([4] * dim)
    state = eq.get_initial_condition(ScalarField.random_uniform(grid))
    field = eq.evolution_rate(state)
    assert field.grid == grid
    rhs = eq._make_pde_rhs_numba(state)
    np.testing.assert_allclose(field.data, rhs(state.data, 0))
    



'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import numpy as np

from .. import cylindrical as ops
from ... import CylindricalGrid, CartesianGrid
from ....fields import ScalarField, VectorField



def test_laplace_cyl():
    """ test the implementation of the laplace operator """
    for boundary_z in ['periodic', 'no-flux']:
        grid = CylindricalGrid(4, (0, 5), (8, 16),
                               periodic_z=(boundary_z == 'periodic'))
        a_2d = np.random.uniform(0, 1, grid.shape)

        bcs = grid.get_boundary_conditions(['no-flux', boundary_z])
        lap_2d = ops.make_laplace(bcs)
        b_2d = lap_2d(a_2d)
        assert b_2d.shape == grid.shape
        
        
        
def test_laplacian_field_cyl():
    """ test the gradient operator """
    grid = CylindricalGrid(2 * np.pi, [0, 2 * np.pi], [8, 16],
                           periodic_z=True)
    r, z = grid.cell_coords[..., 0], grid.cell_coords[..., 1]
    s = ScalarField(grid, data=np.cos(r) + np.sin(z))
    s_lap = s.laplace(bc='natural')
    assert s_lap.data.shape == (8, 16)
    res = -np.cos(r) - np.sin(r) / r - np.sin(z)
    np.testing.assert_allclose(s_lap.data, res, rtol=0.1, atol=0.1)



def test_gradient_field_cyl():
    """ test the gradient operator"""
    grid = CylindricalGrid(2 * np.pi, [0, 2 * np.pi], [8, 16],
                           periodic_z=True)
    r, z = grid.cell_coords[..., 0], grid.cell_coords[..., 1]
    s = ScalarField(grid, data=np.cos(r) + np.sin(z))
    v = s.gradient(bc='natural')
    assert v.data.shape == (3, 8, 16)
    np.testing.assert_allclose(v.data[0], -np.sin(r), rtol=0.1, atol=0.1)
    np.testing.assert_allclose(v.data[1], np.cos(z), rtol=0.1, atol=0.1)
    np.testing.assert_allclose(v.data[2], 0, rtol=0.1, atol=0.1)
     
     
     
def test_divergence_field_cyl():
    """ test the divergence operator """
    grid = CylindricalGrid(2 * np.pi, [0, 2 * np.pi], [8, 16],
                           periodic_z=True)
    r, z = grid.cell_coords[..., 0], grid.cell_coords[..., 1]
    data = [np.cos(r) + np.sin(z)**2,
            np.cos(r)**2 + np.sin(z),
            np.zeros_like(r)]
    v = VectorField(grid, data=data)
    s = v.divergence(bc='natural')
    assert s.data.shape == (8, 16)
    res = np.cos(z) - np.sin(r) + (np.cos(r) + np.sin(z)**2) / r
    np.testing.assert_allclose(s.data, res, rtol=0.1, atol=0.1)
     
     
    
def test_vector_gradient_divergence_field_cyl():
    """ test the divergence operator """
    grid = CylindricalGrid(2 * np.pi, [0, 2 * np.pi], [8, 16],
                           periodic_z=True)
    r, z = grid.cell_coords[..., 0], grid.cell_coords[..., 1]
    data = [np.cos(r) + np.sin(z)**2,
            np.cos(r)**2 + np.sin(z),
            np.zeros_like(r)]
    v = VectorField(grid, data=data)
    t = v.gradient(bc='natural')
    assert t.data.shape == (3, 3, 8, 16)
    v = t.divergence(bc='natural')
    assert v.data.shape == (3, 8, 16)
    
    

def test_findiff_cyl():
    """ test operator for a simple cylindrical grid. Note that we only
    really test the polar symmetry """
    grid = CylindricalGrid(1.5, [0, 1], (3, 2), periodic_z=True)
    _, r1, r2 = grid.axes_coords[0]
    np.testing.assert_array_equal(grid.discretization, np.full(2, 0.5))
    s = ScalarField(grid, [[1, 1], [2, 2], [4, 4]]) 
    v = VectorField(grid, [[[1, 1], [2, 2], [4, 4]],
                           [[0, 0]]*3, [[0, 0]]*3]) 

    # test gradient        
    grad = s.gradient(bc=['value', 'periodic'])
    np.testing.assert_allclose(grad.data[0], [[1, 1], [3, 3], [-6, -6]])
    grad = s.gradient(bc=['derivative', 'periodic'])
    np.testing.assert_allclose(grad.data[0], [[1, 1], [3, 3], [2, 2]])

    # test divergence        
    div = v.divergence(bc=['value', 'periodic'])
    y1 = 3 + 2/r1
    y2 = -6 + 4/r2
    np.testing.assert_allclose(div.data, [[5, 5], [y1, y1], [y2, y2]])
    div = v.divergence(bc=['derivative', 'periodic'])
    y2 = 2 + 4/r2
    np.testing.assert_allclose(div.data, [[5, 5], [y1, y1], [y2, y2]])
    
    # test laplace
    lap = s.laplace(bc=[{'type': 'value', 'value': 3}, 'periodic'])
    y1 = 4 + 3/r1
    y2 = -16
    np.testing.assert_allclose(lap.data, [[8, 8], [y1, y1], [y2, y2]])
    lap = s.laplace(bc=[{'type': 'derivative', 'value': 3}, 'periodic'])
    y2 = -2 + 3.5/r2
    np.testing.assert_allclose(lap.data, [[8, 8], [y1, y1], [y2, y2]])
    
    
    
def test_grid_laplace():
    """ test the cylindrical implementation of the laplace operator """
    grid_cyl = CylindricalGrid(6, (0, 4), (4, 4))
    grid_cart = CartesianGrid([[-5, 5], [-5, 5], [0, 4]], [10, 10, 4]) 

    a_2d = ScalarField.from_expression(grid_cyl,
                                       expression='exp(-5 * r) * cos(z / 3)')     
    a_3d = a_2d.interpolate_to_grid(grid_cart)

    b_2d = a_2d.laplace('natural')
    b_3d = a_3d.laplace('natural')
    b_2d_3 = b_2d.interpolate_to_grid(grid_cart)
     
    np.testing.assert_allclose(b_2d_3.data, b_3d.data, rtol=0.2, atol=0.2)
     
     
     
def test_grid_div_grad():
    """ compare div grad to laplacian """
    grid = CylindricalGrid(2*np.pi, (0, 2*np.pi), (16, 16), periodic_z=True)
    r, z = grid.cell_coords[..., 0], grid.cell_coords[..., 1]
    arr = np.cos(r) + np.sin(z)

    bcs = grid.get_boundary_conditions()
    laplace = grid.get_operator('laplace', bcs)
    grad = grid.get_operator('gradient', bcs)
    div = grid.get_operator('divergence', bcs.differentiated)
    a = laplace(arr)
    b = div(grad(arr))
    res = (-np.sin(r) / r - np.cos(r)) - np.sin(z) 
    # do not test the radial boundary points
    np.testing.assert_allclose(a[1:-1], res[1:-1], rtol=0.1, atol=0.05)
    np.testing.assert_allclose(b[1:-1], res[1:-1], rtol=0.1, atol=0.05)



"""
Classes for tracking simulation results in controlled intervals

Trackers are classes that periodically receive the state of the simulation to
analyze, store, or output it.  The trackers defined in this module are:

.. autosummary::
   :nosignatures:

   ~trackers.CallbackTracker
   ~trackers.ProgressTracker
   ~trackers.PrintTracker
   ~trackers.PlotTracker
   ~trackers.DataTracker
   ~trackers.SteadyStateTracker
   ~trackers.RuntimeTracker
   ~trackers.ConsistencyTracker
   
Multiple trackers can be collected in a :class:`~base.TrackerCollection`,
which provides methods for handling them efficiently.

For each tracker, the interval at which it is called can be decided using one
of the following classes:

.. autosummary::
   :nosignatures:

   ~intervals.ConstantIntervals
   ~intervals.LogarithmicIntervals
   ~intervals.RealtimeIntervals
   
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""


from .intervals import (ConstantIntervals, LogarithmicIntervals,
                        RealtimeIntervals)
from .trackers import *



'''
Cartesian grids of arbitrary dimension.

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
 
'''

import itertools
from abc import ABCMeta
from typing import List, Sequence, Dict, Any, Union, Generator, TYPE_CHECKING

import numpy as np


from .base import GridBase, _check_shape, DimensionError
from ..tools.cuboid import Cuboid
from ..tools.docstrings import fill_in_docstring
from ..tools.plotting import plot_on_axes


if TYPE_CHECKING:
    from .boundaries import Boundaries  # @UnusedImport
    


class CartesianGridBase(GridBase,  # lgtm [py/missing-equals]
                        metaclass=ABCMeta):
    """Base class for :class:`UnitGrid` and :class:`CartesianGrid`"""
    
    cuboid: Cuboid
    
    
    def __init__(self, shape, periodic: Union[List[bool], bool] = False):
        """ 
        Args:
            shape (tuple): The number of support points for each axis. The
                dimension of the grid is given by `len(shape)`.
            periodic (bool or list): Specifies which axes possess periodic
                boundary conditions. This is either a list of booleans defining
                periodicity for each individual axis or a single boolean value
                specifying the same periodicity for all axes.
        """
        super().__init__()
        self._shape = _check_shape(shape)
        self.dim = len(self.shape)
        self.num_axes = self.dim
        
        if isinstance(periodic, (bool, np.bool_)):
            self.periodic = [bool(periodic)] * self.dim
        elif len(periodic) != self.dim:
            raise DimensionError('Number of axes with specified periodicity '
                                 'does not match grid dimension '
                                 f'({len(periodic)} != {self.dim})')
        else:
            self.periodic = periodic
        
        if self.dim <= 3:
            self.axes = list('xyz'[:self.dim])
        else:
            self.axes = [chr(97 + i) for i in range(self.dim)]
        

    @property
    def cell_volume_data(self):
        """ size associated with each cell """
        return tuple(self.discretization)
        

    def contains_point(self, point):
        """ check whether the point is contained in the grid
        
        Args:
            point (vector): Coordinates of the point
        """
        if len(point) != self.dim:
            raise DimensionError('Incompatible dimensions')
        return self.cuboid.contains_point(point)


    def iter_mirror_points(self, point, with_self: bool = False,
                           only_periodic: bool = True) -> Generator:
        """ generates all mirror points corresponding to `point`
        
        Args:
            point (:class:`numpy.ndarray`): the point within the grid
            with_self (bool): whether to include the point itself
            only_periodic (bool): whether to only mirror along periodic axes
        
        Returns:
            A generator yielding the coordinates that correspond to mirrors
        """
        point = np.asanyarray(point, dtype=np.double)
        
        # find all offsets of the individual axes
        offsets = []
        for i in range(self.dim):
            if only_periodic and not self.periodic[i]:
                offsets.append([0])
            else:
                s = self.cuboid.size[i]
                offsets.append([-s, 0, s])
               
        # produce the respective mirrored points 
        for offset in itertools.product(*offsets):
            if with_self or np.linalg.norm(offset) != 0:
                yield point + offset
    
    
    def get_random_point(self, boundary_distance: float = 0,
                         cartesian: bool = True):
        """ return a random point within the grid
        
        Args:
            boundary_distance (float): The minimal distance this point needs to
                have from all boundaries.
            cartesian (bool): Determines whether the point is returned in
                Cartesian coordinates or grid coordinates. This does not have
                any effect for Cartesian coordinate systems, but the argument is
                retained to have a unified interface for all grids.
                
        Returns:
            :class:`numpy.ndarray`: The coordinates of the point        
        """
        # handle the boundary distance
        cuboid = self.cuboid
        if boundary_distance != 0:
            if any(cuboid.size <= 2 * boundary_distance):
                raise RuntimeError('Random points would be too close to '
                                   'boundary')
            cuboid = cuboid.buffer(-boundary_distance)
            
        # create random point
        return cuboid.pos + np.random.random(self.dim) * cuboid.size


    def get_line_data(self, data: np.ndarray, extract: str = 'auto') \
            -> Dict[str, Any]:
        """ return a line cut through the given data
        
        Args:
            data (:class:`numpy.ndarray`):
                The values at the grid points
            extract (str):
                Determines which cut is done through the grid. Possible choices
                are (default is `cut_0`):
                
                * `cut_#`: return values along the axis specified by # and use
                  the mid point along all other axes.
                * `project_#`: average values for all axes, except axis #.

                Here, # can either be a zero-based index (from 0 to dim-1) or
                a letter denoting the axis.
            
        Returns:
            A dictionary with information about the line cut, which is 
            convenient for plotting.
        """
        if data.shape[-self.dim:] != self.shape:
            raise ValueError(f'Shape {data.shape} of the data array is not '
                             f'compatible with grid shape {self.shape}')
            
        def _get_axis(axis):
            """ determine the axis from a given specifier """
            try:
                axis = int(axis)
            except ValueError:
                try:
                    axis = self.axes.index(axis)
                except ValueError:
                    raise ValueError(f'Axis `{axis}` not defined')
            return axis
            
        if extract == 'auto':
            extract = 'cut_0'  # use a cut along first axis
            
        if extract.startswith('cut_'):
            # consider a cut along a given axis
            axis = _get_axis(extract[4:])
            data_y = data
            rank = data.ndim - self.dim  # rank of data 
            for ax in reversed(range(self.dim)):
                if ax != axis:
                    mid_point = self.shape[ax] // 2
                    data_y = np.take(data_y, mid_point, axis=ax + rank)
            label_y = f'Cut along {self.axes[axis]}'
                
        elif extract.startswith('project_'):
            # consider a projection along a given axis
            axis = _get_axis(extract[8:])
            avg_axes = [ax - self.dim
                        for ax in range(self.dim)
                        if ax != axis]
            data_y = data.mean(axis=tuple(avg_axes))
            label_y = f'Projection onto {self.axes[axis]}'
                
        else:
            raise ValueError(f'Unknown extraction method `{extract}`')

        if self.dim == 1:
            label_y = ''

        # return the data with the respective labels                
        return {'data_x': self.axes_coords[axis],
                'data_y': data_y,
                'extent_x': self.axes_bounds[axis],
                'label_x': self.axes[axis],
                'label_y': label_y}    

        
    def get_image_data(self, data) -> Dict[str, Any]:
        """ return a 2d-image of the data
        
        Args:
            data (:class:`numpy.ndarray`): The values at the grid points
            
        Returns:
            A dictionary with information about the image, which is  convenient
            for plotting.
        """
        if data.shape[-self.dim:] != self.shape:
            raise ValueError(f'Shape {data.shape} of the data array is not '
                             f'compatible with grid shape {self.shape}')

        if self.dim == 2:
            image_data = data.T
        elif self.dim == 3: 
            image_data = data[:, :, self.shape[-1] // 2].T
        else:
            raise NotImplementedError('Creating images is only implemented for '
                                      '2d and 3d grids')
            
        extent: List[float] = []
        for c in self.axes_bounds[:2]:
            extent.extend(c)
                
        return {'data': image_data,
                'x': self.axes_coords[-2],
                'y': self.axes_coords[-1],
                'extent': extent,
                'label_x': 'x', 'label_y': 'y'}    


    def point_to_cartesian(self, points):
        """ convert coordinates of a point to Cartesian coordinates
        
        Args:
            points (:class:`numpy.ndarray`):
                Points given in the coordinates of \the grid
                
        Returns:
            :class:`numpy.ndarray`: The Cartesian coordinates of the point
        """
        return points
    
    
    def point_from_cartesian(self, coords):
        """ convert points given in Cartesian coordinates to this grid
        
        Args:
            points (:class:`numpy.ndarray`): Points in Cartesian coordinates.
                
        Returns:
            :class:`numpy.ndarray`: Points given in the coordinates of the grid
        """
        return coords

    
    def polar_coordinates_real(self, origin, ret_angle: bool = False):
        """ return polar coordinates associated with the grid
        
        Args:
            origin (vector): Coordinates of the origin at which the polar
                coordinate system is anchored.
            ret_angle (bool): Determines whether angles are returned alongside
                the distance. If `False` only the distance to the origin is 
                returned for each support point of the grid.
                If `True`, the distance and angles are returned. For a 1d system
                system, the angle is defined as the sign of the difference
                between the point and the origin, so that angles can either be
                1 or -1. For 2d systems and 3d systems, polar coordinates and
                spherical coordinates are used, respectively. 
        """
        origin = np.array(origin, dtype=np.double, ndmin=1)
        if len(origin) != self.dim:
            raise DimensionError('Dimensions are not compatible')
        
        # calculate the difference vector between all cells and the origin
        diff = self.difference_vector_real(origin, self.cell_coords)
        dist = np.linalg.norm(diff, axis=-1)  # get distance
        
        # determine distance and optionally angles for these vectors
        if ret_angle:
            if self.dim == 1:
                return dist, np.sign(diff)[..., 0]
            
            elif self.dim == 2:
                return dist, np.arctan2(diff[:, :, 0], diff[:, :, 1])
            
            elif self.dim == 3:
                theta = np.arccos(diff[..., 2] / dist)
                phi = np.arctan2(diff[..., 0], diff[..., 1])
                return dist, theta, phi 
            
            else:
                raise NotImplementedError('Cannot calculate angles for '
                                          f'dimension {self.dim}')
        else:
            return dist
        
        
    def from_polar_coordinates(self, distance, angle, origin=None):
        """ convert polar coordinates to Cartesian coordinates
        
        This function is currently only implemented for 1d and 2d systems.
        
        Args:
            distance: The radial distance
            angle: The angle with respect to the origin
            origin (optional): Sets the origin of the coordinate system. If
                omitted. The zero point is assumed as the origin.
                
        Returns:
            The Cartesian coordinates corresponding to the given polar
            coordinates. 
        """
        distance = np.asarray(distance)
        angle = np.asarray(angle)
        if origin is None:
            origin = np.zeros(self.dim)
        else:
            origin = np.atleast_1d(origin)
        
        if self.dim == 1:
            diff = distance * angle
            coords = origin + diff[..., None]
        
        elif self.dim == 2:
            unit_vector = np.moveaxis([np.sin(angle), np.cos(angle)], 0, -1)
            diff = distance[..., None] * unit_vector
            coords = origin + diff
        
        else:
            raise NotImplementedError('Cannot calculate coordinates for '
                                      f'dimension {self.dim}')
            
        return self.normalize_point(coords)


    @plot_on_axes()
    def plot(self, ax, **kwargs):
        r""" visualize the grid
        
        Args:
            ax (:class:`matplotlib.axes.Axes`):
                Figure axes to be used for plotting. If `None`, a new figure is
                created. This has no effect if a `reference` is supplied.
            \**kwargs: Extra arguments are passed on the to the matplotlib
                plotting routines, e.g., to set the color of the lines
        """
        if self.dim not in {1, 2}:
            raise NotImplementedError('Plotting is not implemented for grids '
                                      f'of dimension {self.dim}')
            
        kwargs.setdefault('color', 'k')
        xb = self.axes_bounds[0]
        for x in np.linspace(*xb, self.shape[0] + 1):
            ax.axvline(x, **kwargs)
        ax.set_xlim(*xb)
        ax.set_xlabel(self.axes[0])
        
        if self.dim == 2:
            yb = self.axes_bounds[1]
            for y in np.linspace(*yb, self.shape[1] + 1):
                ax.axhline(y, **kwargs)
            ax.set_ylim(*yb)
            ax.set_ylabel(self.axes[1])
            
            ax.set_aspect(1)
            

    @fill_in_docstring
    def get_boundary_conditions(self, bc='natural', rank: int = 0) \
            -> "Boundaries":
        """ constructs boundary conditions from a flexible data format
        
        Args:
            bc (str or list or tuple or dict):
                The boundary conditions applied to the field. 
                {ARG_BOUNDARIES}  
            rank (int):
                The tensorial rank of the value associated with the boundary
                conditions.

        Raises:
            ValueError: If the data given in `bc` cannot be read
            PeriodicityError: If the boundaries are not compatible with the
                periodic axes of the grid. 
        """            
        from .boundaries import Boundaries  # @Reimport
        return Boundaries.from_data(self, bc, rank=rank)
    


class UnitGrid(CartesianGridBase):
    r""" d-dimensional Cartesian grid with unit discretization in all directions 
    
    The grids can be thought of as a collection of d-dimensional cells of unit
    length. The `shape` parameter determines how many boxes there are in each
    direction. The cells are enumerated starting with 0, so the last cell has
    index :math:`n-1` if there are :math:`n` cells along a dimension. A given
    cell :math:`i` extends from coordinates :math:`i` to :math:`i + 1`, so the
    midpoint is at :math:`i + \frac12`, which is the cell coordinate. Taken
    together, the cells covers the interval :math:`[0, n]` along this dimension.
    """
    
    
    def __init__(self, shape, periodic: Union[List[bool], bool] = False):
        """ 
        Args:
            shape (tuple): The number of support points for each axis. The
                dimension of the grid is given by `len(shape)`.
            periodic (bool or list): Specifies which axes possess periodic
                boundary conditions. This is either a list of booleans defining
                periodicity for each individual axis or a single boolean value
                specifying the same periodicity for all axes.
        """
        super().__init__(shape, periodic)
        self.cuboid = Cuboid(np.zeros(self.dim), self.shape)
        self._discretization = np.ones(self.dim)
        
        # determine the cell center coordinates
        self._axes_coords = tuple(np.arange(n) + 0.5 for n in self.shape)
        self._axes_bounds = tuple(self.cuboid.bounds)


    @property
    def state(self) -> Dict[str, Any]:
        """ dict: the state of the grid """
        return {'shape': self.shape,
                'periodic': self.periodic}


    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "UnitGrid":  # type: ignore
        """ create a field from a stored `state`.
        
        Args:
            state (dict):
                The state from which the grid is reconstructed.
        """
        state_copy = state.copy()
        obj = cls(shape=state_copy.pop('shape'),
                  periodic=state_copy.pop('periodic'))
        if state_copy:
            raise ValueError(f'State items {state_copy.keys()} were not used')
        return obj


    @property
    def volume(self) -> float:
        """ float: total volume of the grid """
        return float(np.prod(self.shape))


    def normalize_point(self, point, reduced_coords: bool = False):
        """ normalize coordinates by applying periodic boundary conditions
        
        Args:
            points (:class:`numpy.ndarray`): An array of coordinates
            reduced_coords (bool): Flag determining whether only the coordinates
                corresponding to axes in this grid are given
                 
        Returns:
            :class:`numpy.ndarray`: The respective coordinates with periodic
            boundary conditions applied.
        """
        point = np.asarray(point, dtype=np.double)
        if point.size == 0:
            return np.zeros((0, self.dim))
        if point.ndim == 0:
            if self.dim > 1:
                raise DimensionError(f'Dimension mismatch: Point {point} is '
                                     f'not of dimension {self.dim}.')
        elif point.shape[-1] != self.dim:
            raise DimensionError('Dimension mismatch: Array of shape '
                                 f'{point.shape} does not describe points of '
                                 f'dimension {self.dim}.')
        
        # normalize the coordinates for the periodic dimensions
        if point.ndim == 0:
            if self.periodic[0]:
                point %= self.shape[0]
        else:
            for i in range(self.num_axes):
                if self.periodic[i]:
                    point[..., i] %= self.shape[i]

        return point
    

    def cell_to_point(self, cells, cartesian: bool = True):
        """ convert cell coordinates to real coordinates
        
        Args:
            cells (:class:`numpy.ndarray`): Indices of the cells whose center
                coordinates are requested. This can be float values to indicate
                positions relative to the cell center.
            cartesian (bool): Determines whether the point is returned in
                Cartesian coordinates or grid coordinates. This does not have
                any effect for Cartesian coordinate systems, but the argument is
                retained to have a unified interface for all grids.
                
        Returns:
            :class:`numpy.ndarray`: The center points of the respective cells
        """
        cells = np.asanyarray(cells, dtype=np.double)
        if cells.size == 0:
            return np.zeros((0, self.dim))
        if cells.shape[-1] != self.dim:
            raise DimensionError('Dimension mismatch')
        return cells + 0.5


    def point_to_cell(self, points):
        """ Determine cell(s) corresponding to given point(s)

        This function respects periodic boundary conditions, but it does not
        throw an error when coordinates lie outside the bcs (for
        non-periodic axes).
        
        Args:
            points (:class:`numpy.ndarray`): Real coordinates
                
        Returns:
            :class:`numpy.ndarray`: The indices of the respective cells
        """
        return self.normalize_point(points).astype(np.int)
        
    
    def difference_vector_real(self, p1, p2):
        """ return the vector pointing from p1 to p2.
        
        In case of periodic boundary conditions, the shortest vector is returned
        
        Args:
            p1 (:class:`numpy.ndarray`): First point(s)
            p2 (:class:`numpy.ndarray`): Second point(s)
            
        Returns:
            :class:`numpy.ndarray`: The difference vectors between the points
            with periodic boundary conditions applied.
        """
        diff = np.atleast_1d(p2) - np.atleast_1d(p1)
        # correct the periodic dimensions
        for i in range(self.num_axes):
            if self.periodic[i]:
                s = self.shape[i]
                diff[..., i] = (diff[..., i] + s/2) % s - s/2
        return diff
    
    
    def to_cartesian(self) -> "CartesianGrid":
        """ convert unit grid to CartesianGrid """
        return CartesianGrid(self.cuboid.bounds, shape=self.shape,
                             periodic=self.periodic)
        
        
    def get_subgrid(self, indices: Sequence[int]) -> "UnitGrid":
        """ return a subgrid of only the specified axes
        
        Args:
            indices (list):
                Indices indicating the axes that are retained in the subgrid
                
        Returns:
            :class:`UnitGrid`: The subgrid
        """
        subgrid = self.__class__(shape=[self.shape[i] for i in indices],
                                 periodic=[self.periodic[i] for i in indices])
        subgrid.axes = [self.axes[i] for i in indices]
        return subgrid



class CartesianGrid(CartesianGridBase):
    r""" d-dimensional Cartesian grid with uniform discretization for each axis
    
    The grids can be thought of as a collection of n-dimensional boxes, called
    cells, of equal length in each dimension. The bounds then defined the total
    volume covered by these cells, while the cell coordinates give the location
    of the box centers. We index the boxes starting from 0 along each dimension.
    Consequently, the cell :math:`i-\frac12` corresponds to the left edge of the 
    covered interval and the index :math:`i+\frac12` corresponds to the right
    edge, when the dimension is covered by d boxes.
    
    In particular, the discretization along dimension :math:`k` is defined as

    .. math::
            x^{(k)}_i &= x^{(k)}_\mathrm{min} + \left(i + \frac12\right)
                \Delta x^{(k)}
            \quad \text{for} \quad i = 0, \ldots, N^{(k)} - 1
        \\
            \Delta x^{(k)} &= \frac{x^{(k)}_\mathrm{max} -
                                    x^{(k)}_\mathrm{min}}{N^{(k)}}
                                    
    where :math:`N^{(k)}` is the number of cells along this dimension.
    Consequently, the cells have dimension :math:`\Delta x^{(k)}` and cover the
    interval :math:`[x^{(k)}_\mathrm{min}, x^{(k)}_\mathrm{max}]`.
    """
    
    
    def __init__(self, bounds, shape,
                 periodic: Union[List[bool], bool] = False):
        """ 
        Args:
            bounds (tuple): Give the coordinate range for each axis. This should
                be a tuple of two number (lower and upper bound) for each axis.
                The length of `bounds` thus determines the grid dimension. 
            shape (tuple): The number of support points for each axis. The
                length of `shape` needs to match the grid dimension. 
            periodic (bool or list): Specifies which axes possess periodic
                boundary conditions. This is either a list of booleans defining
                periodicity for each individual axis or a single boolean value
                specifying the same periodicity for all axes.
        """
        bounds = np.array(bounds, ndmin=1, dtype=np.double)
        if bounds.shape == (2,):
            raise ValueError('`bounds with shape (2,) are ambiguous. Either '
                             'use shape (1, 2) to set up a 1d system with two '
                             'bounds or shape (2, 1) for a 2d system with only '
                             'the upper bounds specified') 
        
        if bounds.ndim == 1 or bounds.shape[1] == 1:
            # only set the upper bounds
            bounds = np.atleast_1d(np.squeeze(bounds))
            self.cuboid = Cuboid(np.zeros_like(bounds), bounds, mutable=False)
            
        elif bounds.ndim == 2 and bounds.shape[1] == 2:
            # upper and lower bounds of the grid are given
            self.cuboid = Cuboid.from_bounds(bounds, mutable=False)
            
        else:
            raise ValueError('Do not know how to interpret shape '
                             f'{bounds.shape} for bounds')
            
        # handle the shape array
        shape = _check_shape(shape)
        if len(shape) == 1 and self.cuboid.dim > 1:
            shape = np.full(self.cuboid.dim, shape, dtype=np.uint32)
        if self.cuboid.dim != len(shape):
            raise DimensionError('Dimension of the bounds and the shape are '
                                 'not compatible')
                    
        # initialize the base class            
        super().__init__(shape, periodic)
            
        # determine the coordinates
        p1, p2 = self.cuboid.corners
        axes_coords, discretization = [], []
        for d in range(self.dim):
            c, dc = np.linspace(p1[d], p2[d], self.shape[d],
                                endpoint=False, retstep=True)
            if self.shape[d] == 1:
                # correct for singular dimension
                dc = p2[d] - p1[d]
            c += dc / 2
            axes_coords.append(c)
            discretization.append(dc)
        self._discretization = np.array(discretization)
        self._axes_coords = tuple(axes_coords)
        self._axes_bounds = tuple(self.cuboid.bounds)

    
    @property
    def state(self) -> Dict[str, Any]:
        """ dict: the state of the grid """
        return {'bounds': self.axes_bounds,
                'shape': self.shape,
                'periodic': self.periodic}


    @classmethod
    def from_state(cls,  # type: ignore
                   state: Dict[str, Any]) -> "CartesianGrid":
        """ create a field from a stored `state`.
        
        Args:
            state (dict):
                The state from which the grid is reconstructed.
        """
        state_copy = state.copy()
        obj = cls(bounds=state_copy.pop('bounds'),
                  shape=state_copy.pop('shape'),
                  periodic=state_copy.pop('periodic'))
        if state_copy:
            raise ValueError(f'State items {state_copy.keys()} were not used')
        return obj

                    
    @property
    def volume(self) -> float:
        """ float: total volume of the grid """
        return float(self.cuboid.volume)


    def normalize_point(self, point, reduced_coords: bool = False):
        """ normalize coordinates by applying periodic boundary conditions
        
        Args:
            points (:class:`numpy.ndarray`): An array of coordinates
            reduced_coords (bool): Flag determining whether only the coordinates
                corresponding to axes in this grid are given
                 
        Returns:
            :class:`numpy.ndarray`: The respective coordinates with periodic
            boundary conditions applied.
        """
        point = np.asarray(point, dtype=np.double)
        if point.size == 0:
            return np.zeros((0, self.dim)) 
        if point.ndim == 0:
            if self.dim > 1:
                raise DimensionError(f'Dimension mismatch: Point {point} is '
                                     f'not of dimension {self.dim}.')
        elif point.shape[-1] != self.dim:
            raise DimensionError('Dimension mismatch: Array of shape '
                                 f'{point.shape} does not describe points of '
                                 f'dimension {self.dim}.')
        
        if any(self.periodic):
            if self.dim == 1:  # single, periodic dimension
                offset = self.cuboid.pos
                point = (point - offset) % self.cuboid.size + offset
            else:  # multiple dimensions => extract the periodic ones
                offset = self.cuboid.pos[self.periodic]
                size = self.cuboid.size[self.periodic]
                point[..., self.periodic] = \
                            (point[..., self.periodic] - offset) % size + offset
        return point


    def cell_to_point(self, cells, cartesian: bool = True):
        """ convert cell coordinates to real coordinates
        
        Args:
            cells (:class:`numpy.ndarray`): Indices of the cells whose center
                coordinates are requested. This can be float values to indicate
                positions relative to the cell center.
            cartesian (bool): Determines whether the point is returned in
                Cartesian coordinates or grid coordinates. This does not have
                any effect for Cartesian coordinate systems, but the argument is
                retained to have a unified interface for all grids.
                
        Returns:
            :class:`numpy.ndarray`: The center points of the respective cells
        """
        cells = np.atleast_1d(cells)
        if cells.size == 0:
            return cells
        else:
            return self.cuboid.pos + (cells + 0.5) * self.discretization


    def point_to_cell(self, points):
        """ Determine cell(s) corresponding to given point(s)

        This function respects periodic boundary conditions, but it does not
        throw an error when coordinates lie outside the bcs (for
        non-periodic axes).
        
        Args:
            points (:class:`numpy.ndarray`): Real coordinates
                
        Returns:
            :class:`numpy.ndarray`: The indices of the respective cells
        """
        points = self.normalize_point(points)
        cell_coords = (points - self.cuboid.pos) / self.discretization
        return cell_coords.astype(np.int)

        
    def difference_vector_real(self, p1, p2):
        """ return the vector pointing from p1 to p2.
        
        In case of periodic boundary conditions, the shortest vector is returned
        
        Args:
            p1 (:class:`numpy.ndarray`): First point(s)
            p2 (:class:`numpy.ndarray`): Second point(s)
            
        Returns:
            :class:`numpy.ndarray`: The difference vectors between the points
            with periodic  boundary conditions applied.
        """
        diff = np.atleast_1d(p2) - np.atleast_1d(p1)
        if any(self.periodic):
            size = self.cuboid.size[self.periodic]
            diff[..., self.periodic] = \
                            (diff[..., self.periodic] + size/2) % size - size/2
        return diff


    def get_subgrid(self, indices: Sequence[int]) -> "CartesianGrid":
        """ return a subgrid of only the specified axes
        
        Args:
            indices (list):
                Indices indicating the axes that are retained in the subgrid
                
        Returns:
            :class:`CartesianGrid`: The subgrid
        """
        subgrid = self.__class__(bounds=[self.axes_bounds[i] for i in indices],
                                 shape=[self.shape[i] for i in indices],
                                 periodic=[self.periodic[i] for i in indices])
        subgrid.axes = [self.axes[i] for i in indices]
        return subgrid

